__all__ = ('GuildDiscovery', )

from scarletio import RichAttributeErrorBaseType

from ..discovery_category import DiscoveryCategory

from .fields import (
    parse_application_actioned, parse_application_requested, parse_emoji_discovery, parse_keywords,
    parse_primary_category, parse_sub_categories, put_application_actioned, put_application_requested,
    put_emoji_discovery, put_keywords, put_primary_category, put_sub_categories,
    validate_application_actioned, validate_application_requested, validate_emoji_discovery, validate_keywords,
    validate_primary_category, validate_sub_categories
)


class GuildDiscovery(RichAttributeErrorBaseType):
    """
    Represent a guild's Discovery settings.
    
    Attributes
    ----------
    application_actioned : `None | DateTime`
        When the guild's application was accepted or rejected.
    application_requested : `None | DateTime`
        When the guild applied to guild discovery. Only set if pending.
    emoji_discovery : `bool`
        Whether guild info is shown when the respective guild's emojis are clicked.
    keywords : `None | tuple<str>`
        The set discovery search keywords for the guild.
    primary_category : ``DiscoveryCategory``
        The primary discovery category of the guild.
    sub_categories : `None`, `tuple` of ``DiscoveryCategory``
        Guild Discovery sub-categories. Up to 5.
    """
    __slots__ = (
        'application_actioned', 'application_requested', 'emoji_discovery', 'keywords', 'primary_category',
        'sub_categories'
    )
    
    def __new__(
        cls,
        *,
        application_actioned = ...,
        application_requested = ...,
        emoji_discovery = ...,
        keywords = ...,
        primary_category = ...,
        sub_categories = ...,
    ):
        """
        Creates a new guild discovery from the given fields.
        
        Parameters
        ----------
        application_actioned : `None | DateTime`, Optional (Keyword only)
            When the guild's application was accepted or rejected.
        application_requested : `None | DateTime`, Optional (Keyword only)
            When the guild applied to guild discovery. Only set if pending.
        emoji_discovery : `bool`, Optional (Keyword only)
            Whether guild info is shown when the respective guild's emojis are clicked.
        keywords : `None`, `iterable` of `str`, Optional (Keyword only)
            The set discovery search keywords for the guild.
        primary_category : ``DiscoveryCategory``, `int`, Optional (Keyword only)
            The primary discovery category of the guild.
        sub_categories : `None`, `iterable` of (`int`, ``DiscoveryCategory``), Optional (Keyword only)
            Guild Discovery sub-categories.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # application_actioned
        if application_actioned is ...:
            application_actioned = None
        else:
            application_actioned = validate_application_actioned(application_actioned)
        
        # application_requested
        if application_requested is ...:
            application_requested = None
        else:
            application_requested = validate_application_requested(application_requested)
        
        # emoji_discovery
        if emoji_discovery is ...:
            emoji_discovery = False
        else:
            emoji_discovery = validate_emoji_discovery(emoji_discovery)
        
        # keywords
        if keywords is ...:
            keywords = None
        else:
            keywords = validate_keywords(keywords)
        
        # primary_category
        if primary_category is ...:
            primary_category = DiscoveryCategory.general
        else:
            primary_category = validate_primary_category(primary_category)
        
        # sub_categories
        if sub_categories is ...:
            sub_categories = None
        else:
            sub_categories = validate_sub_categories(sub_categories)
        
        # Construct object
        
        self = object.__new__(cls)
        self.application_actioned = application_actioned
        self.application_requested = application_requested
        self.emoji_discovery = emoji_discovery
        self.keywords = keywords
        self.primary_category = primary_category
        self.sub_categories = sub_categories
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new guild discovery object from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Guild discovery data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.application_actioned = parse_application_actioned(data)
        self.application_requested = parse_application_requested(data)
        self.emoji_discovery = parse_emoji_discovery(data)
        self.keywords = parse_keywords(data)
        self.primary_category = parse_primary_category(data)
        self.sub_categories = parse_sub_categories(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the guild discovery to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        if include_internals:
            put_application_actioned(self.application_actioned, data, defaults)
            put_application_requested(self.application_requested, data, defaults)
        
        put_emoji_discovery(self.emoji_discovery, data, defaults)
        put_keywords(self.keywords, data, defaults)
        put_primary_category(self.primary_category, data, defaults)
        put_sub_categories(self.sub_categories, data, defaults)
        
        return data
    
    
    def __eq__(self, other):
        """Returns whether the two guild discoveries are the same."""
        if (type(self) is not type(other)):
            return NotImplemented
        
        # application_actioned
        if (self.application_actioned != other.application_actioned):
            return False
        
        # application_requested
        if (self.application_requested != other.application_requested):
            return False
        
        # emoji_discovery
        if (self.emoji_discovery != other.emoji_discovery):
            return False
        
        # keywords
        if (self.keywords != other.keywords):
            return False
        
        # primary_category
        if (self.primary_category is not other.primary_category):
            return False
        
        # sub_categories
        if (self.sub_categories != other.sub_categories):
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the guild discovery's hash value."""
        hash_value = 0
        
        # application_actioned
        application_actioned = self.application_actioned
        if (application_actioned is not None):
            hash_value ^= hash(application_actioned)
        
        # application_requested
        application_requested = self.application_requested
        if (application_requested is not None):
            hash_value ^= hash(application_requested)
        
        # emoji_discovery
        emoji_discovery = self.emoji_discovery
        if (emoji_discovery is not None):
            hash_value ^= emoji_discovery
        
        # keywords
        keywords = self.keywords
        if (keywords is not None):
            hash_value ^= len(keywords) << 1
            
            for keyword in keywords:
                hash_value ^= hash(keyword)
        
        # primary_category
        hash_value ^= self.primary_category.value << 5
        
        # sub_categories
        sub_categories = self.sub_categories
        if (sub_categories is not None):
            hash_value ^= len(sub_categories) << 9
            
            shift = 13
            
            for sub_category in sub_categories:
                hash_value ^= sub_category.value << shift
                shift += 4
        
        return hash_value
    
    
    def __repr__(self):
        """Returns the guild discovery's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        field_added = False
        
        # application_actioned
        application_actioned = self.application_actioned
        if (application_actioned is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' application_actioned = ')
            repr_parts.append(repr(application_actioned))
        
        # application_requested
        application_requested = self.application_requested
        if (application_requested is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' application_requested = ')
            repr_parts.append(repr(application_requested))
        
        # emoji_discovery
        emoji_discovery = self.emoji_discovery
        if emoji_discovery:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' emoji_discovery = ')
            repr_parts.append(repr(emoji_discovery))
        
        # keywords
        keywords = self.keywords
        if (keywords is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' keywords = [')
            
            index = 0
            length = len(keywords)
            
            while True:
                keyword = keywords[index]
                repr_parts.append(repr(keyword))
                
                index += 1
                if index == length:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        # primary_category
        primary_category = self.primary_category
        if (primary_category is not DiscoveryCategory.general):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' primary_category = ')
            repr_parts.append(repr(primary_category.name))
            repr_parts.append(' ~ ')
            repr_parts.append(repr(primary_category.value))
        
        # sub_categories
        sub_categories = self.sub_categories
        if (sub_categories is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' sub_categories = [')
            
            index = 0
            length = len(sub_categories)
            
            while True:
                sub_category = sub_categories[index]
                repr_parts.append(repr(sub_category.name))
                repr_parts.append(' ~ ')
                repr_parts.append(repr(sub_category.value))
                
                index += 1
                if index == length:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
            
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def copy(self):
        """
        Copies the guild discovery.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.application_actioned = self.application_actioned
        new.application_requested = self.application_requested
        new.emoji_discovery = self.emoji_discovery
        keywords = self.keywords
        if (keywords is not None):
            keywords = (*keywords,)
        new.keywords = keywords
        new.primary_category = self.primary_category
        sub_categories = self.sub_categories
        if (sub_categories is not None):
            sub_categories = (*sub_categories,)
        new.sub_categories = sub_categories
        return new
    
    
    def copy_with(
        self,
        *,
        application_actioned = ...,
        application_requested = ...,
        emoji_discovery = ...,
        keywords = ...,
        primary_category = ...,
        sub_categories = ...,
    ):
        """
        Copies the guild discovery with the given fields.
        
        Parameters
        ----------
        application_actioned : `None | DateTime`, Optional (Keyword only)
            When the guild's application was accepted or rejected.
        application_requested : `None | DateTime`, Optional (Keyword only)
            When the guild applied to guild discovery. Only set if pending.
        emoji_discovery : `bool`, Optional (Keyword only)
            Whether guild info is shown when the respective guild's emojis are clicked.
        keywords : `None`, `iterable` of `str`, Optional (Keyword only)
            The set discovery search keywords for the guild.
        primary_category : ``DiscoveryCategory``, `int`, Optional (Keyword only)
            The primary discovery category of the guild.
        sub_categories : `None`, `iterable` of (`int`, ``DiscoveryCategory``), Optional (Keyword only)
            Guild Discovery sub-categories.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # application_actioned
        if application_actioned is ...:
            application_actioned = self.application_actioned
        else:
            application_actioned = validate_application_actioned(application_actioned)
        
        # application_requested
        if application_requested is ...:
            application_requested = self.application_requested
        else:
            application_requested = validate_application_requested(application_requested)
        
        # emoji_discovery
        if emoji_discovery is ...:
            emoji_discovery = self.emoji_discovery
        else:
            emoji_discovery = validate_emoji_discovery(emoji_discovery)
        
        # keywords
        if keywords is ...:
            keywords = self.keywords
            if (keywords is not None):
                keywords = (*keywords,)
        else:
            keywords = validate_keywords(keywords)
        
        # primary_category
        if primary_category is ...:
            primary_category = self.primary_category
        else:
            primary_category = validate_primary_category(primary_category)
        
        # sub_categories
        if sub_categories is ...:
            sub_categories = self.sub_categories
            if (sub_categories is not None):
                sub_categories = (*sub_categories,)
        else:
            sub_categories = validate_sub_categories(sub_categories)
        
        # Construct object
        
        new = object.__new__(type(self))
        new.application_actioned = application_actioned
        new.application_requested = application_requested
        new.emoji_discovery = emoji_discovery
        new.keywords = keywords
        new.primary_category = primary_category
        new.sub_categories = sub_categories
        return new
    
    
    def iter_keywords(self):
        """
        Iterates over the keywords of the guild discovery.
        
        This method is an iterable generator.
        
        Yields
        ------
        keyword : `str`
        """
        keywords = self.keywords
        if (keywords is not None):
            yield from keywords
    
    
    def iter_sub_categories(self):
        """
        Iterates over the sub-categories of the guild discovery.
        
        This method is an iterable generator.
        
        Yields
        ------
        sub_category : ``DiscoveryCategory``
        """
        sub_categories = self.sub_categories
        if (sub_categories is not None):
            yield from sub_categories



__all__ = ('EmbedMediaFlag',)

from ...bases import FlagBase, FlagDescriptor as F


class EmbedMediaFlag(FlagBase):
    """
    Bitwise flags of an embed media.
    
    The implemented embed media flags are the following:
    
    +-------------------+-------------------+
    | Respective name   | Bitwise position  |
    +===================+===================+
    | animated          | 5                 |
    +-------------------+-------------------+
    """
    # 0 ??
    # 1 ??
    # 2 ??
    # 3 ??
    # 4 ??
    animated = F(5)



__all__ = ()

from scarletio import un_map_pack

from ...discord.application_command import ApplicationCommandOptionType

from ...discord.channel import Channel, ChannelType
from ...discord.client import Client
from ...discord.interaction import InteractionEvent
from ...discord.message import Attachment
from ...discord.role import Role
from ...discord.user import ClientUserBase, User, UserBase

ANNOTATION_TYPE_STR = 1
ANNOTATION_TYPE_INT = 2
ANNOTATION_TYPE_BOOL = 3
ANNOTATION_TYPE_USER = 4
ANNOTATION_TYPE_USER_ID = 5
ANNOTATION_TYPE_ROLE = 6
ANNOTATION_TYPE_ROLE_ID = 7
ANNOTATION_TYPE_CHANNEL = 8
ANNOTATION_TYPE_CHANNEL_ID = 9
ANNOTATION_TYPE_NUMBER = 10
ANNOTATION_TYPE_MENTIONABLE = 11
ANNOTATION_TYPE_MENTIONABLE_ID = 12
ANNOTATION_TYPE_SELF_CLIENT = 13
ANNOTATION_TYPE_SELF_INTERACTION_EVENT = 14
ANNOTATION_TYPE_EXPRESSION = 15
ANNOTATION_TYPE_FLOAT = 16
ANNOTATION_TYPE_SELF_TARGET = 17
ANNOTATION_TYPE_SELF_VALUE = 18
ANNOTATION_TYPE_ATTACHMENT = 19

ANNOTATION_NAMES_CLIENT = frozenset((
    'c',
    'client',
))

ANNOTATION_NAMES_INTERACTION_EVENT = frozenset((
    'e',
    'event',
    'interaction_event',
))

ANNOTATION_NAMES_TARGET = frozenset((
    't',
    'target',
))


ANNOTATION_NAMES_VALUE = frozenset((
    'v',
    'val',
    'value',
))

CHANNEL_TYPES_FORBIDDEN = (
    ChannelType.unknown,
    ChannelType.thread,
)

CHANNEL_TYPES_GROUP_GUILD = tuple(
    channel_type for channel_type in ChannelType.INSTANCES.values()
    if (channel_type not in CHANNEL_TYPES_FORBIDDEN) and channel_type.flags.guild
)
CHANNEL_TYPES_GUILD_CATEGORY = (ChannelType.guild_category, )
CHANNEL_TYPES_GUILD_DIRECTORY = (ChannelType.guild_directory, )
CHANNEL_TYPES_GUILD_STORE = (ChannelType.guild_store, )
CHANNEL_TYPES_GUILD_SYSTEM = tuple(
    channel_type for channel_type in ChannelType.INSTANCES.values()
    if (channel_type not in CHANNEL_TYPES_FORBIDDEN) and channel_type.flags.guild_system
)
CHANNEL_TYPES_GUILD_TEXT = (ChannelType.guild_text, )
CHANNEL_TYPES_GUILD_ANNOUNCEMENTS = (ChannelType.guild_announcements, )
CHANNEL_TYPES_GUILD_CONNECTABLE = tuple(
    channel_type for channel_type in ChannelType.INSTANCES.values()
    if (channel_type not in CHANNEL_TYPES_FORBIDDEN) and channel_type.flags.guild and channel_type.flags.connectable
)
CHANNEL_TYPES_GUILD_VOICE = (ChannelType.guild_voice, )
CHANNEL_TYPES_GUILD_STAGE = (ChannelType.guild_stage, )
CHANNEL_TYPES_GROUP_PRIVATE = tuple(
    channel_type for channel_type in ChannelType.INSTANCES.values()
    if (channel_type not in CHANNEL_TYPES_FORBIDDEN) and channel_type.flags.private
)
CHANNEL_TYPES_PRIVATE = (ChannelType.private, )
CHANNEL_TYPES_PRIVATE_GROUP = (ChannelType.private_group, )
CHANNEL_TYPES_GROUP_THREAD = tuple(
    channel_type for channel_type in ChannelType.INSTANCES.values()
    if (channel_type not in CHANNEL_TYPES_FORBIDDEN) and channel_type.flags.thread
)
CHANNEL_TYPES_THREAD_ANNOUNCEMENTS = (ChannelType.guild_thread_announcements, )
CHANNEL_TYPES_THREAD_PUBLIC = (ChannelType.guild_thread_public, )
CHANNEL_TYPES_THREAD_PRIVATE = (ChannelType.guild_thread_private, )
CHANNEL_TYPES_GROUP_TEXTUAL = tuple(
    channel_type for channel_type in ChannelType.INSTANCES.values()
    if (channel_type not in CHANNEL_TYPES_FORBIDDEN) and channel_type.flags.textual
)
CHANNEL_TYPES_GROUP_GUILD_TEXTUAL = tuple(
    channel_type for channel_type in ChannelType.INSTANCES.values()
    if (channel_type not in CHANNEL_TYPES_FORBIDDEN) and channel_type.flags.guild and channel_type.flags.textual
)
CHANNEL_TYPES_GROUP_CONNECTABLE = tuple(
    channel_type for channel_type in ChannelType.INSTANCES.values()
    if (channel_type not in CHANNEL_TYPES_FORBIDDEN) and channel_type.flags.connectable
)
CHANNEL_TYPES_GROUP_THREADABLE = tuple(
    channel_type for channel_type in ChannelType.INSTANCES.values()
    if (channel_type not in CHANNEL_TYPES_FORBIDDEN) and channel_type.flags.threadable
)
CHANNEL_TYPES_GUILD_FORUM = (ChannelType.guild_forum, )
CHANNEL_TYPES_GUILD_MEDIA = (ChannelType.guild_media, )

CHANNEL_TYPES_GROUP_FORUM = tuple(
    channel_type for channel_type in ChannelType.INSTANCES.values()
    if (channel_type not in CHANNEL_TYPES_FORBIDDEN) and channel_type.flags.forum
)

CHANNEL_TYPES_GROUP_INVITABLE = tuple(
    channel_type for channel_type in ChannelType.INSTANCES.values()
    if (channel_type not in CHANNEL_TYPES_FORBIDDEN) and channel_type.flags.invitable
)

STR_ANNOTATION_TO_ANNOTATION_TYPE = {
    # Generic
    'str': (ANNOTATION_TYPE_STR, None),
    'int': (ANNOTATION_TYPE_INT, None),
    'bool': (ANNOTATION_TYPE_BOOL, None),
    'user': (ANNOTATION_TYPE_USER, None),
    'user_id': (ANNOTATION_TYPE_USER_ID, None),
    'role': (ANNOTATION_TYPE_ROLE, None),
    'role_id': (ANNOTATION_TYPE_ROLE_ID, None),
    'channel': (ANNOTATION_TYPE_CHANNEL, None),
    'channel_id': (ANNOTATION_TYPE_CHANNEL_ID, None),
    'number': (ANNOTATION_TYPE_NUMBER, None),
    'mentionable': (ANNOTATION_TYPE_MENTIONABLE, None),
    'mentionable_id': (ANNOTATION_TYPE_MENTIONABLE_ID, None),
    'expression': (ANNOTATION_TYPE_EXPRESSION, None),
    'float': (ANNOTATION_TYPE_FLOAT, None),
    'attachment': (ANNOTATION_TYPE_ATTACHMENT, None),
    
    # Channel type specific
    # - by channel name
    f'{"channel"}{"guild"}{"base"}': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GROUP_GUILD),
    f'{"channel"}{"guild"}{"main"}{"base"}': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GROUP_GUILD),
    f'{"channel"}{"category"}': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_CATEGORY),
    f'{"channel"}{"directory"}': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_DIRECTORY),
    f'{"channel"}{"store"}': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_STORE),
    f'{"channel"}{"text"}': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_SYSTEM),
    f'{"channel"}{"voice"}{"base"}': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_CONNECTABLE),
    f'{"channel"}{"voice"}': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_VOICE),
    f'{"channel"}{"stage"}': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_STAGE),
    f'{"channel"}{"private"}': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_PRIVATE),
    f'{"channel"}{"group"}': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_PRIVATE_GROUP),
    f'{"channel"}{"thread"}': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GROUP_THREAD),
    f'{"channel"}{"text"}{"base"}': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GROUP_GUILD_TEXTUAL),
    f'{"channel"}{"forum"}':(ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_FORUM),
    f'{"channel"}{"media"}':(ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_MEDIA),
    f'{"channel"}{"forum"}{"base"}':(ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GROUP_FORUM),
    # - by generic name
    'channel_guild_text': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_TEXT),
    'channel_private': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_PRIVATE),
    'channel_guild_voice': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_VOICE),
    'channel_private_group': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_PRIVATE_GROUP),
    'channel_guild_category': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_CATEGORY),
    'channel_guild_announcements': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_ANNOUNCEMENTS),
    'channel_guild_store': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_STORE),
    'channel_guild_thread_announcements': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_THREAD_ANNOUNCEMENTS),
    'channel_guild_thread_public': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_THREAD_PUBLIC),
    'channel_guild_thread_private': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_THREAD_PRIVATE),
    'channel_guild_stage': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_STAGE),
    'channel_guild_directory': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_DIRECTORY),
    'channel_group_messageable': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GROUP_TEXTUAL),
    'channel_group_guild_messageable': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GROUP_GUILD_TEXTUAL),
    'channel_group_guild_main_text': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_SYSTEM),
    'channel_group_connectable': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GROUP_CONNECTABLE),
    'channel_group_private': (ANNOTATION_TYPE_CHANNEL, ),
    'channel_group_guild_connectable': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_CONNECTABLE),
    'channel_group_guild': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GROUP_GUILD),
    'channel_group_thread': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GROUP_THREAD),
    'channel_group_can_contain_threads': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GROUP_THREADABLE),
    'channel_guild_forum': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_FORUM),
    'channel_guild_media': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GUILD_MEDIA),
    'channel_group_forum': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GROUP_FORUM),
    'channel_group_invitable': (ANNOTATION_TYPE_CHANNEL, CHANNEL_TYPES_GROUP_INVITABLE),
    # - id + by generic name
    'channel_id_guild_text': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GUILD_TEXT),
    'channel_id_private': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_PRIVATE),
    'channel_id_guild_voice': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GUILD_VOICE),
    'channel_id_private_group': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_PRIVATE_GROUP),
    'channel_id_guild_category': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GUILD_CATEGORY),
    'channel_id_guild_announcements': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GUILD_ANNOUNCEMENTS),
    'channel_id_guild_store': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GUILD_STORE),
    'channel_id_guild_thread_announcements': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_THREAD_ANNOUNCEMENTS),
    'channel_id_guild_thread_public': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_THREAD_PUBLIC),
    'channel_id_guild_thread_private': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_THREAD_PRIVATE),
    'channel_id_guild_stage': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GUILD_STAGE),
    'channel_id_guild_directory': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GUILD_DIRECTORY),
    'channel_id_group_messageable': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GROUP_TEXTUAL),
    'channel_id_group_guild_messageable': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GROUP_GUILD_TEXTUAL),
    'channel_id_group_guild_main_text': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GUILD_SYSTEM),
    'channel_id_group_connectable': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GROUP_CONNECTABLE),
    'channel_id_group_private': (ANNOTATION_TYPE_CHANNEL_ID, ),
    'channel_id_group_guild_connectable': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GUILD_CONNECTABLE),
    'channel_id_group_guild': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GROUP_GUILD),
    'channel_id_group_thread': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GROUP_THREAD),
    'channel_id_group_can_contain_threads': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GROUP_THREADABLE),
    'channel_id_guild_forum': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GUILD_FORUM),
    'channel_id_guild_media': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GUILD_MEDIA),
    'channel_id_group_forum': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GROUP_FORUM),
    'channel_id_group_invitable': (ANNOTATION_TYPE_CHANNEL_ID, CHANNEL_TYPES_GROUP_INVITABLE),
    
    # Internal
    **un_map_pack((
        (name, (ANNOTATION_TYPE_SELF_CLIENT, None))
        for name in ANNOTATION_NAMES_CLIENT
    )),
    **un_map_pack((
        (name, (ANNOTATION_TYPE_SELF_INTERACTION_EVENT, None))
        for name in ANNOTATION_NAMES_INTERACTION_EVENT
    )),
    **un_map_pack((
        (name, (ANNOTATION_TYPE_SELF_TARGET, None))
        for name in ANNOTATION_NAMES_TARGET
    )),
    **un_map_pack((
        (name, (ANNOTATION_TYPE_SELF_VALUE, None))
        for name in ANNOTATION_NAMES_VALUE
    )),
}

# Used at repr
ANNOTATION_TYPE_TO_STR_ANNOTATION = {
    ANNOTATION_TYPE_STR: 'str',
    ANNOTATION_TYPE_INT: 'int',
    ANNOTATION_TYPE_BOOL: 'bool',
    ANNOTATION_TYPE_USER: 'user',
    ANNOTATION_TYPE_USER_ID: 'user_id',
    ANNOTATION_TYPE_ROLE: 'role',
    ANNOTATION_TYPE_ROLE_ID: 'role_id',
    ANNOTATION_TYPE_CHANNEL: 'channel',
    ANNOTATION_TYPE_CHANNEL_ID: 'channel_id',
    ANNOTATION_TYPE_NUMBER: 'number',
    ANNOTATION_TYPE_MENTIONABLE: 'mentionable',
    ANNOTATION_TYPE_MENTIONABLE_ID : 'mentionable_id',
    ANNOTATION_TYPE_EXPRESSION: 'expression',
    ANNOTATION_TYPE_FLOAT: 'float',
    ANNOTATION_TYPE_ATTACHMENT : 'attachment',
    
    ANNOTATION_TYPE_SELF_CLIENT: 'client',
    ANNOTATION_TYPE_SELF_INTERACTION_EVENT: 'interaction_event',
    ANNOTATION_TYPE_SELF_TARGET: 'target',
    ANNOTATION_TYPE_SELF_VALUE: 'value',
}

TYPE_ANNOTATION_TO_ANNOTATION_TYPE = {
    # Generic
    str: (ANNOTATION_TYPE_STR, None),
    int: (ANNOTATION_TYPE_INT, None),
    bool: (ANNOTATION_TYPE_BOOL, None),
    ClientUserBase: (ANNOTATION_TYPE_USER, None),
    UserBase: (ANNOTATION_TYPE_USER, None),
    User: (ANNOTATION_TYPE_USER, None),
    Role: (ANNOTATION_TYPE_ROLE, None),
    Channel: (ANNOTATION_TYPE_CHANNEL, None),
    float: (ANNOTATION_TYPE_FLOAT, None),
    Attachment: (ANNOTATION_TYPE_ATTACHMENT, None),
    
    # Internal
    Client: (ANNOTATION_TYPE_SELF_CLIENT, None),
    InteractionEvent: (ANNOTATION_TYPE_SELF_INTERACTION_EVENT, None),
}

INTERNAL_ANNOTATION_TYPES = frozenset((
    ANNOTATION_TYPE_SELF_CLIENT,
    ANNOTATION_TYPE_SELF_INTERACTION_EVENT,
))

# `int` Discord fields are broken and they are refusing to fix it, use string instead.
# Reference: https://github.com/discord/discord-api-docs/issues/2448
ANNOTATION_TYPE_TO_OPTION_TYPE = {
    ANNOTATION_TYPE_STR: ApplicationCommandOptionType.string,
    ANNOTATION_TYPE_INT: ApplicationCommandOptionType.string,
    ANNOTATION_TYPE_BOOL: ApplicationCommandOptionType.boolean,
    ANNOTATION_TYPE_USER: ApplicationCommandOptionType.user,
    ANNOTATION_TYPE_USER_ID: ApplicationCommandOptionType.user,
    ANNOTATION_TYPE_ROLE: ApplicationCommandOptionType.role,
    ANNOTATION_TYPE_ROLE_ID: ApplicationCommandOptionType.role,
    ANNOTATION_TYPE_CHANNEL: ApplicationCommandOptionType.channel,
    ANNOTATION_TYPE_CHANNEL_ID: ApplicationCommandOptionType.channel,
    ANNOTATION_TYPE_NUMBER: ApplicationCommandOptionType.integer,
    ANNOTATION_TYPE_MENTIONABLE: ApplicationCommandOptionType.mentionable,
    ANNOTATION_TYPE_MENTIONABLE_ID: ApplicationCommandOptionType.mentionable,
    ANNOTATION_TYPE_EXPRESSION: ApplicationCommandOptionType.string,
    ANNOTATION_TYPE_FLOAT: ApplicationCommandOptionType.float,
    ANNOTATION_TYPE_ATTACHMENT: ApplicationCommandOptionType.attachment,
    
    ANNOTATION_TYPE_SELF_CLIENT: ApplicationCommandOptionType.none,
    ANNOTATION_TYPE_SELF_INTERACTION_EVENT: ApplicationCommandOptionType.none,
    ANNOTATION_TYPE_SELF_TARGET: ApplicationCommandOptionType.none,
    ANNOTATION_TYPE_SELF_VALUE: ApplicationCommandOptionType.none,
}

ANNOTATION_TYPE_TO_REPRESENTATION = {
    ANNOTATION_TYPE_STR: 'string',
    ANNOTATION_TYPE_INT: 'integer',
    ANNOTATION_TYPE_BOOL: 'bool',
    ANNOTATION_TYPE_USER: 'user',
    ANNOTATION_TYPE_USER_ID: 'user',
    ANNOTATION_TYPE_ROLE: 'role',
    ANNOTATION_TYPE_ROLE_ID: 'role',
    ANNOTATION_TYPE_CHANNEL: 'channel',
    ANNOTATION_TYPE_CHANNEL_ID: 'channel',
    ANNOTATION_TYPE_NUMBER: 'integer',
    ANNOTATION_TYPE_MENTIONABLE: 'mentionable',
    ANNOTATION_TYPE_MENTIONABLE_ID : 'mentionable',
    ANNOTATION_TYPE_EXPRESSION: 'expression',
    ANNOTATION_TYPE_FLOAT: 'float',
    ANNOTATION_TYPE_ATTACHMENT: 'attachment',
}


ANNOTATION_AUTO_COMPLETE_AVAILABILITY = frozenset((
    ANNOTATION_TYPE_STR,
    ANNOTATION_TYPE_INT,
    ANNOTATION_TYPE_EXPRESSION,
    ANNOTATION_TYPE_NUMBER,
    ANNOTATION_TYPE_FLOAT,
))




__all__ = ()

from ...onboarding import OnboardingPromptOption, OnboardingPromptType
from ...onboarding.onboarding_prompt.fields import (
    validate_in_onboarding, validate_name, validate_options, validate_required, validate_single_select, validate_type
)

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup
from ..conversion_helpers.converters import value_deserializer_name, value_serializer_name


# ---- in_onboarding ----

IN_ONBOARDING_CONVERSION = AuditLogEntryChangeConversion(
    ('in_onboarding',),
    'in_onboarding',
    value_validator = validate_in_onboarding,
)


@IN_ONBOARDING_CONVERSION.set_value_deserializer
def in_onboarding_value_deserializer(value):
    if value is None:
        value = False
    return value


# ---- name ----

NAME_CONVERSION = AuditLogEntryChangeConversion(
    ('title',),
    'name',
    value_deserializer = value_deserializer_name,
    value_serializer = value_serializer_name,
    value_validator = validate_name,
)


# ---- options ----

OPTIONS_CONVERSION = AuditLogEntryChangeConversion(
    ('options',),
    'options',
    value_validator = validate_options,
)


@OPTIONS_CONVERSION.set_value_deserializer
def options_value_deserializer(value):
    if value is None:
        pass
    elif (not value):
        value = None
    else:
        value = (*(OnboardingPromptOption.from_data(data) for data in value),)
    
    return value


@OPTIONS_CONVERSION.set_value_serializer
def options_value_serializer(value):
    if value is None:
        value = []
    else:
        value = [option.to_data(defaults = True, include_internals = True) for option in value]
    
    return value


# ---- required ----

REQUIRED_CONVERSION = AuditLogEntryChangeConversion(
    ('required',),
    'required',
    value_validator = validate_required,
)


@REQUIRED_CONVERSION.set_value_deserializer
def required_value_deserializer(value):
    if value is None:
        value = False
    return value


# ---- single_select ----

SINGLE_SELECT_CONVERSION = AuditLogEntryChangeConversion(
    ('single_select',),
    'single_select',
    value_validator = validate_single_select,
)


@SINGLE_SELECT_CONVERSION.set_value_deserializer
def single_select_value_deserializer(value):
    if value is None:
        value = False
    return value


# ---- type ----

TYPE_CONVERSION = AuditLogEntryChangeConversion(
    ('type',),
    'type',
    value_validator = validate_type,
)


@TYPE_CONVERSION.set_value_deserializer
def type_value_deserializer(value):
    return OnboardingPromptType(value)


@TYPE_CONVERSION.set_value_serializer
def type_value_serializer(value):
    return value.value


# ---- Construct ----

ONBOARDING_PROMPT_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    IN_ONBOARDING_CONVERSION,
    NAME_CONVERSION,
    OPTIONS_CONVERSION,
    REQUIRED_CONVERSION,
    SINGLE_SELECT_CONVERSION,
    TYPE_CONVERSION,
)



__all__ = ('Sticker', )

from ...bases import DiscordEntity
from ...core import GUILDS, STICKERS, STICKER_PACKS
from ...http.urls import build_sticker_url, build_sticker_url_as
from ...precreate_helpers import process_precreate_parameters_and_raise_extra
from ...user import ZEROUSER
from ...utils import DATETIME_FORMAT_CODE

from .constants import SORT_VALUE_DEFAULT
from .fields import (
    parse_available, parse_description, parse_format, parse_guild_id, parse_id, parse_name, parse_pack_id,
    parse_sort_value, parse_tags, parse_type, parse_user, put_available, put_description, put_format,
    put_guild_id, put_id, put_name, put_pack_id, put_sort_value, put_tags, put_type,
    put_user, validate_available, validate_description, validate_format, validate_guild_id, validate_id,
    validate_name, validate_pack_id, validate_sort_value, validate_tags, validate_type, validate_user
)
from .preinstanced import StickerFormat, StickerType


PRECREATE_FIELDS = {
    'description': ('description', validate_description),
    'available': ('available', validate_available),
    'guild': ('guild_id', validate_guild_id),
    'guild_id': ('guild_id', validate_guild_id),
    'name': ('name', validate_name),
    'pack': ('pack_id', validate_pack_id),
    'pack_id': ('pack_id', validate_pack_id),
    'sticker_format': ('format', validate_format),
    'sticker_type': ('type', validate_type),
    'sort_value': ('sort_value', validate_sort_value),
    'tags': ('tags', validate_tags),
    'user': ('user', validate_user),
}

STICKER_TYPE_NONE = StickerType.none
STICKER_TYPE_STANDARD = StickerType.standard
STICKER_TYPE_GUILD = StickerType.guild


class Sticker(DiscordEntity, immortal = True):
    """
    Represents a ``Message``'s sticker.
    
    Attributes
    ----------
    available : `bool`
        Whether the sticker is available.
    description : `None`, `str`
        The sticker's description.
    format : ``StickerFormat``
        The sticker's format.
    guild_id : `int`
        The guild's identifier to what the sticker is bound to. Defaults to `0` if the sticker is not bound to any
        guild.
    id : `int`
        The unique identifier number of the sticker.
    name : `str`
        The sticker's name.
    pack_id : `int`
        The unique identifier number of the pack from the sticker is. Defaults to `0` if the sticker is not bound to
        any pack.
    sort_value : `int`
        Value used to sort the stickers.
    tags : None`, `frozenset` of `str`
        Tags of the sticker if applicable.
    type : ``StickerType``
        The sticker's type.
    user : ``ClientUserBase``
        The user who uploaded the sticker. Defaults to ``ZEROUSER``.
    """
    __slots__ = (
        'available', 'description', 'format', 'guild_id', 'name', 'pack_id', 'sort_value', 'tags', 'type', 'user'
    )
    
    def __new__(
        cls,
        *,
        available = ...,
        description = ...,
        name = ...,
        pack_id = ...,
        sort_value = ...,
        sticker_format = ...,
        sticker_type = ...,
        tags = ...,
        user = ...,
    ):
        """
        Creates a partial sticker with the given fields.
        
        Parameters
        ----------
        available : `bool`, Optional (Keyword only)
            Whether the sticker is available.
        description : `None`, `str`, Optional (Keyword only)
            The sticker's description.
        name : `str`, Optional (Keyword only)
            The sticker's name.
        pack_id : `int`, ``StickerPack``, Optional (Keyword only)
            The unique identifier number of the pack from the sticker is.
        sort_value : `int`, Optional (Keyword only)
            Value used to sort the stickers.
        sticker_format : ``StickerFormat``, `int`, Optional (Keyword only)
            The sticker's format.
        sticker_type : ``StickerType``, `int`, Optional (Keyword only)
            The sticker's type.
        tags : None`, `iterable` of `str`, Optional (Keyword only)
            Tags of the sticker if applicable.
        user : ``ClientUserBase``, Optional (Keyword only)
            The user who uploaded the sticker.
        
        Raises
        ------
        TypeError
            - A parameter's type is incorrect.
        ValueError
            - A parameter's value is incorrect.
        """
        # available
        if available is ...:
            available = True
        else:
            available = validate_available(available)
        
        # description
        if description is ...:
            description = None
        else:
            description = validate_description(description)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # pack_id
        if pack_id is ...:
            pack_id = 0
        else:
            pack_id = validate_pack_id(pack_id)
        
        # sort_value
        if sort_value is ...:
            sort_value = SORT_VALUE_DEFAULT
        else:
            sort_value = validate_sort_value(sort_value)
        
        # sticker_format
        if sticker_format is ...:
            sticker_format = StickerFormat.none
        else:
            sticker_format = validate_format(sticker_format)
        
        # sticker_type
        if sticker_type is ...:
            sticker_type = StickerType.none
        else:
            sticker_type = validate_type(sticker_type)
        
        # tags
        if tags is ...:
            tags = None
        else:
            tags = validate_tags(tags)
        
        # user
        if user is ...:
            user = ZEROUSER
        else:
            user = validate_user(user)
    
        # Construct
        self = object.__new__(cls)
        self.available = available
        self.description = description
        self.format = sticker_format
        self.guild_id = 0
        self.id = 0
        self.name = name
        self.pack_id = pack_id
        self.sort_value = sort_value
        self.tags = tags
        self.type = sticker_type
        self.user = user
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new sticker from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Sticker data.
        
        Returns
        -------
        sticker : `instance<cls>`
        """
        sticker_id = parse_id(data)
        
        try:
            self = STICKERS[sticker_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = sticker_id
            self._set_attributes(data)
            
            STICKERS[sticker_id] = self
            
        else:
            if not self.partial:
                # Set user if received
                user = parse_user(data)
                if user is not ZEROUSER:
                    self.user = user
                
                return self
            
            self._set_attributes(data)
        
        # Do not register, since that ruins `client.events.sticker__create` after a `client.sticker_create` call.
        # guild_id = self.guild_id
        # if guild_id:
        #     try:
        #         guild = GUILDS[guild_id]
        #     except KeyError:
        #         pass
        #     else:
        #         guild.stickers[sticker_id] = self
        
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the sticker to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_description(self.description, data, defaults)
        put_name(self.name, data, defaults)
        put_tags(self.tags, data, defaults)
        
        if include_internals:
            put_available(self.available, data, defaults)
            put_sort_value(self.sort_value, data, defaults)
            put_format(self.format, data, defaults)
            put_guild_id(self.guild_id, data, defaults)
            put_id(self.id, data, defaults)
            put_pack_id(self.pack_id, data, defaults)
            put_sort_value(self.sort_value, data, defaults)
            put_type(self.type, data, defaults)
            put_user(self.user, data, defaults)
        
        return data
    
    
    @classmethod
    def precreate(cls, sticker_id, **keyword_parameters):
        """
        Precreates the sticker by creating a partial one with the given parameters. When the sticker is loaded
        the precreated one will be picked up. If an already existing sticker would be precreated, returns that
        instead and updates that only, if that is partial.
        
        Parameters
        ----------
        sticker_id : `int`
            The sticker's identifier.
        **keyword_parameters : keyword parameters
            Additional predefined attributes for the sticker.
        
        Other Parameters
        ----------------
        available : `bool`, Optional (Keyword only)
            Whether the sticker is available.
        description : `None`, `str`, Optional (Keyword only)
            The sticker's description.
        guild : ``int | Guild``, Optional (Keyword only)
            Alternative for `guild_id`.
        guild_id : ``int | Guild``, Optional (Keyword only)
             The sticker's guild's identifier.
        name : `str`, Optional (Keyword only)
            The sticker's name.
        pack : `int`, ``StickerPack``, Optional (Keyword only)
            Alternative for `pack_id`.
        pack_id : `int`, ``StickerPack``, Optional (Keyword only)
            The unique identifier number of the pack from the sticker is.
        sort_value : `int`, Optional (Keyword only)
            Value used to sort the stickers.
        sticker_format : ``StickerFormat``, `int`, Optional (Keyword only)
            The sticker's format.
        sticker_type : ``StickerType``, `int`, Optional (Keyword only)
            The sticker's type.
        tags : None`, `iterable` of `str`, Optional (Keyword only)
            Tags of the sticker if applicable.
        user : ``ClientUserBase``, Optional (Keyword only)
            The user who uploaded the sticker.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - Extra parameter(s) given.
            - A parameter's type is incorrect.
        ValueError
            - A parameter's value is incorrect.
        """
        sticker_id = validate_id(sticker_id)

        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        try:
            self = STICKERS[sticker_id]
        except KeyError:
            self = cls._create_empty(sticker_id)
            STICKERS[sticker_id] = self
        else:
            if not self.partial:
                return self
        
        if (processed is not None):
            for name, value in processed:
                setattr(self, name, value)
        
        return self
    
    
    def __repr__(self):
        """Returns the sticker's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        sticker_id = self.id
        if sticker_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(self.id))
            repr_parts.append(',')
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    

    def __format__(self, code):
        """
        Formats the sticker in a format string.
        
        Parameters
        ----------
        code : `str`
            The option on based the result will be formatted.
        
        Returns
        -------
        sticker : `str`
        
        Raises
        ------
        ValueError
            Unknown format code.
        
        Examples
        --------
        ```py
        >>> from hata import Sticker, now_as_id
        >>> sticker = Sticker.precreate(now_as_id(), name = 'nice')
        >>> sticker
        <Sticker id = 712359434843586560, name = 'nice'>
        >>> # no code returns the sticker's name.
        >>> f'{sticker}'
        'nice'
        >>> # 'c' stands for created at.
        >>> f'{sticker:c}'
        '2020.05.19-17:42:04'
        ```
        """
        if not code:
            return self.name
        
        if code == 'c':
            return format(self.created_at, DATETIME_FORMAT_CODE)
        
        raise ValueError(
            f'Unknown format code {code!r} for {self.__class__.__name__}; {self!r}. '
            f'Available format codes: {""!r}, {"c"!r}.'
        )
    
    def __hash__(self):
        """Returns the sticker's hash."""
        sticker_id = self.id
        if sticker_id:
            return sticker_id
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Returns a partial sticker's hash value.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # available
        hash_value ^= self.available
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= hash(description)
        
        # format
        hash_value ^= self.format.value << 1
        
        # name
        name = self.name
        if (description is None) or (description != name):
            hash_value ^= hash(name)
        
        # pack_id
        hash_value ^= self.pack_id
        
        # sort_value
        hash_value ^= self.sort_value << 5
        
        # tags
        tags = self.tags
        if (tags is not None):
            hash_value ^= len(tags) << 9
            
            for tag in tags:
                hash_value ^= hash(tag)
        
        # type
        hash_value ^= self.type.value << 13
        
        # user
        hash_value ^= hash(self.user)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two stickers are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two stickers are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def __gt__(self, other):
        """Returns whether self is greater than other."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_sort_value = self.sort_value
        other_sort_value = other.sort_value
        if self_sort_value > other_sort_value:
            return True
        
        if self_sort_value == other_sort_value:
            if self.id > other.id:
                return True
        
        return False
    
    def __ge__(self, other):
        """Returns whether self is greater or equal to other."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_sort_value = self.sort_value
        other_sort_value = other.sort_value
        if self_sort_value > other_sort_value:
            return True
        
        if self_sort_value == other_sort_value:
            if self.id > other.id:
                return True
        
        return self._is_equal_same_type(other)
    
    
    def __le__(self, other):
        """Returns whether self is less or equal to other."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_sort_value = self.sort_value
        other_sort_value = other.sort_value
        if self_sort_value < other_sort_value:
            return True
        
        if self_sort_value == other_sort_value:
            if self.id < other.id:
                return True
        
        return self._is_equal_same_type(other)
    
    
    def __lt__(self, other):
        """Returns whether self is less than other."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_sort_value = self.sort_value
        other_sort_value = other.sort_value
        if self_sort_value < other_sort_value:
            return True
        
        if self_sort_value == other_sort_value:
            if self.id < other.id:
                return True
        
        return False
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two stickers are equal. `self` and `other` must be the same type.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other sticker.
        
        Returns
        -------
        is_equal : `bool`
        """
        self_id = self.id
        other_id = other.id
        if (self_id and other_id):
            return self_id == other_id
        
        # available
        if (self.available != other.available):
            return False
        
        # description
        if (self.description != other.description):
            return False
        
        # format
        if (self.format is not other.format):
            return False
        
        # guild_id
        # Skip | non-partial
        
        # name
        if (self.name != other.name):
            return False
        
        # pack_id
        if (self.pack_id != other.pack_id):
            return False
        
        # sort_value
        if (self.sort_value != other.sort_value):
            return False
        
        # tags
        if (self.tags != other.tags):
            return False
        
        # type
        if (self.type is not other.type):
            return False
        
        # user
        if (self.user != other.user):
            return False
        
        return True
    

    @property
    def partial(self):
        """
        Returns whether the sticker is partial.
        
        Returns
        -------
        partial : `bool`
        """
        sticker_type = self.type
        if sticker_type is STICKER_TYPE_NONE:
            return True
        
        sticker_id = self.id
        if not sticker_id:
            return True
        
        if sticker_type is STICKER_TYPE_GUILD:
            guild_id = self.guild_id
            if not guild_id:
                return True
            
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                return True
            
            if self.id not in guild.stickers:
                return True
            
            return guild.partial
        
        if sticker_type is STICKER_TYPE_STANDARD:
            pack_id = self.pack_id
            if not pack_id:
                return True
            
            try:
                sticker_pack = STICKER_PACKS[pack_id]
            except KeyError:
                return True
            
            return (not sticker_pack.has_sticker(self))
        
        return True
    
    
    def _set_attributes(self, data):
        """
        Sets the attributes of the sticker from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Sticker data.
        """
        self.format = parse_format(data)
        self.guild_id = parse_guild_id(data)
        self.pack_id = parse_pack_id(data)
        self.type = parse_type(data)
        self.user = ZEROUSER
        
        self._update_attributes(data)
    
    
    def _update_attributes(self, data):
        """
        Updates the stickers with the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Sticker data.
        """
        self.available = parse_available(data)
        self.description = parse_description(data)
        self.name = parse_name(data)
        self.sort_value = parse_sort_value(data)
        self.tags = parse_tags(data)
        
        # set user if applicable
        user = parse_user(data)
        if user is not ZEROUSER:
            self.user = user
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the sticker with the given data and returns the changed attributes in `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Sticker data.
        
        Returns
        -------
        old_attributes : `dict<str, object>`
            All item in the returned dictionary is optional.
            
            +-----------------------+-----------------------------------+
            | Keys                  | Values                            |
            +=======================+===================================+
            | available             | `bool`                            |
            +-----------------------+-----------------------------------+
            | description           | `None`, `str`                     |
            +-----------------------+-----------------------------------+
            | name                  | `str`                             |
            +-----------------------+-----------------------------------+
            | sort_value            | `int`                             |
            +-----------------------+-----------------------------------+
            | tags                  | `None`  or `frozenset` of `str`   |
            +-----------------------+-----------------------------------+
        """
        old_attributes = {}
        
        available = parse_available(data)
        if self.available != available:
            old_attributes['available'] = self.available
            self.available = available
        
        description = parse_description(data)
        if description != self.description:
            old_attributes['description'] = self.description
            self.description = description
        
        name = parse_name(data)
        if name != self.name:
            old_attributes['name'] = self.name
            self.name = name
        
        sort_value = parse_sort_value(data)
        if sort_value != self.sort_value:
            old_attributes['sort_value'] = self.sort_value
            self.sort_value = sort_value
        
        tags = parse_tags(data)
        if tags != self.tags:
            old_attributes['tags'] = self.tags
            self.tags = tags
        
        # set user if applicable
        user = parse_user(data)
        if user is not ZEROUSER:
            self.user = user
        
        return old_attributes
    
    
    @classmethod
    def _create_empty(cls, sticker_id):
        """
        Creates an empty sticker with the given identifier.
        
        Parameters
        ----------
        sticker_id : `int`
            The sticker's identifier.
        
        Returns
        -------
        self : `instance<cls>`
            The created sticker.
        """
        self = object.__new__(cls)
        self.id = sticker_id
        self.available = True
        self.description = None
        self.format = StickerFormat.none
        self.guild_id = 0
        self.name = ''
        self.pack_id = 0
        self.sort_value = SORT_VALUE_DEFAULT
        self.tags = None
        self.type = StickerType.none
        self.user = ZEROUSER
        return self
    
    
    def copy(self):
        """
        Copies the sticker returning a partial one.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.available = self.available
        new.description = self.description
        new.format = self.format
        new.guild_id = 0
        new.id = 0
        new.name = self.name
        new.pack_id = self.pack_id
        new.sort_value = self.sort_value
        tags = self.tags
        if (tags is not None):
            tags = frozenset(iter(tags))
        new.tags = tags
        new.type = self.type
        new.user = self.user
        return new
    
    
    def copy_with(
        self,
        *,
        available = ...,
        description = ...,
        name = ...,
        pack_id = ...,
        sort_value = ...,
        sticker_format = ...,
        sticker_type = ...,
        tags = ...,
        user = ...,
    ):
        """
        Copies the sticker with the given fields.
        
        Parameters
        ----------
        available : `bool`, Optional (Keyword only)
            Whether the sticker is available.
        description : `None`, `str`, Optional (Keyword only)
            The sticker's description.
        name : `str`, Optional (Keyword only)
            The sticker's name.
        pack_id : `int`, ``StickerPack``, Optional (Keyword only)
            The unique identifier number of the pack from the sticker is.
        sort_value : `int`, Optional (Keyword only)
            Value used to sort the stickers.
        sticker_format : ``StickerFormat``, `int`, Optional (Keyword only)
            The sticker's format.
        sticker_type : ``StickerType``, `int`, Optional (Keyword only)
            The sticker's type.
        tags : None`, `iterable` of `str`, Optional (Keyword only)
            Tags of the sticker if applicable.
        user : ``ClientUserBase``, Optional (Keyword only)
            The user who uploaded the sticker.
        
        Raises
        ------
        TypeError
            - A parameter's type is incorrect.
        ValueError
            - A parameter's value is incorrect.
        """
        # available
        if available is ...:
            available = self.available
        else:
            available = validate_available(available)
        
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # pack_id
        if pack_id is ...:
            pack_id = self.pack_id
        else:
            pack_id = validate_pack_id(pack_id)
        
        # sort_value
        if sort_value is ...:
            sort_value = self.sort_value
        else:
            sort_value = validate_sort_value(sort_value)
        
        # sticker_format
        if sticker_format is ...:
            sticker_format = self.format
        else:
            sticker_format = validate_format(sticker_format)
        
        # sticker_type
        if sticker_type is ...:
            sticker_type = self.type
        else:
            sticker_type = validate_type(sticker_type)
        
        # tags
        if tags is ...:
            tags = self.tags
            if (tags is not None):
                tags = frozenset(iter(tags))
        else:
            tags = validate_tags(tags)
        
        # user
        if user is ...:
            user = self.user
        else:
            user = validate_user(user)
        
        # Construct
        new = object.__new__(type(self))
        new.available = available
        new.description = description
        new.format = sticker_format
        new.guild_id = 0
        new.id = 0
        new.name = name
        new.pack_id = pack_id
        new.sort_value = sort_value
        new.tags = tags
        new.type = sticker_type
        new.user = user
        return new
    
    
    @property
    def guild(self):
        """
        Returns the sticker's guild if cached.
        
        Returns
        -------
        guild : ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    def iter_tags(self):
        """
        Iterates over the tags of the sticker.
        
        This method is an iterable generator.
        
        Yields
        ------
        tag : `str`
        """
        tags = self.tags
        if (tags is not None):
            yield from tags
    
    
    def has_tag(self, tag):
        """
        Returns whether the sticker has the given tag.
        
        Parameters
        ----------
        tag : `str`
            The tag to check for.
        
        Returns
        -------
        has_tag : `bool`
        """
        tags = self.tags
        if (tags is None):
            return False
        
        return (tag in self.tags)
    
    
    @property
    def url(self):
        """
        Returns the sticker's url. If the sticker has no url, returns `None`.
        
        Returns
        -------
        url : `None | str`
        """
        return build_sticker_url(self.id, self.format)
    
    
    def url_as(self, size = None, preview = False):
        """
        Returns the sticker's url. If the sticker has no url, returns `None`.
        
        Parameters
        ----------
        size : `None | int` = `None`, Optional
            The preferred minimal size of the icon's url.
        
        preview : `bool` = `False`, Optional
            Whether preview url should be generated.
        
        Returns
        -------
        url : `None | str`
        """
        return build_sticker_url_as(self.id, self.format, size, preview)



__all__ = ('GuildEnhancementEntitlementsCreateEvent',)

from scarletio import copy_docs

from ...bases import EventBase
from ...core import GUILDS

from .fields import (
    parse_guild_id, parse_entitlements, put_guild_id, put_entitlements, validate_guild_id, validate_entitlements
)


class GuildEnhancementEntitlementsCreateEvent(EventBase):
    """
    Represents a `GUILD_POWERUP_ENTITLEMENTS_CREATE` event.
    
    Attributes
    ----------
    entitlements : ``None | tuple<Entitlement>``
        The affected entitlements.
    
    guild_id : `int`
        The guild's identifier where the event is for.
    """
    __slots__ = ('entitlements', 'guild_id')
    
    def __new__(cls, *, entitlements = ..., guild_id = ...):
        """
        Creates a new guild enhancement entitlements create / delete event.
        
        Parameters
        ----------
        entitlements : ``None | iterable<Entitlement>``, Optional (Keyword only)
            The affected entitlements.
        
        guild_id : ``None | int | Guild``, Optional (Keyword only)
            The guild or its identifier where the event will be.
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is invalid.
        """
        # entitlements
        if entitlements is ...:
            entitlements = None
        else:
            entitlements = validate_entitlements(entitlements)
        
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_guild_id(guild_id)
        
        # Construct
        self = object.__new__(cls)
        self.entitlements = entitlements
        self.guild_id = guild_id
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new guild enhancement entitlements create / delete event from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Guild enhancement entitlements create / delete event.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.entitlements = parse_entitlements(data)
        self.guild_id = parse_guild_id(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serializes the guild enhancement entitlements create / delete event.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields of their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_entitlements(self.entitlements, data, defaults)
        put_guild_id(self.guild_id, data, defaults)
        return data
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        # entitlements
        repr_parts.append(' entitlements = ')
        repr_parts.append(repr(self.entitlements))
        
        # guild_id
        repr_parts.append(', guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 2
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.guild_id
        yield self.entitlements
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # entitlements
        if self.entitlements != other.entitlements:
            return False
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # entitlements
        entitlements = self.entitlements
        if (entitlements is not None):
            hash_value ^= len(entitlements)
            for entitlement in entitlements:
                hash_value ^= hash(entitlement)
        
        # guild_id
        hash_value ^= self.guild_id
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the guild enhancement entitlements create / delete event.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        
        # entitlements
        entitlements = self.entitlements
        if (entitlements is not None):
            entitlements = (*(entitlement.copy() for entitlement in entitlements),)
        
        new.entitlements = entitlements
        
        new.guild_id = self.guild_id
        return new
    
    
    def copy_with(self, *, entitlements = ..., guild_id = ...):
        """
        Copies the guild enhancement entitlements create / delete event with the given fields.
        
        Parameters
        ----------
        entitlements : ``None | iterable<Entitlement>``, Optional (Keyword only)
            The affected entitlements.
        
        guild_id : ``None | int | Guild``, Optional (Keyword only)
            The guild or its identifier where the event will be.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is invalid.
        """
        # entitlements
        if entitlements is ...:
            entitlements = self.entitlements
            if (entitlements is not None):
                entitlements = (*(entitlement.copy() for entitlement in entitlements),)
        else:
            entitlements = validate_entitlements(entitlements)
        
        # guild_id
        if guild_id is ...:
            guild_id = self.guild_id
        else:
            guild_id = validate_guild_id(guild_id)
        
        # Construct
        new = object.__new__(type(self))
        new.entitlements = entitlements
        new.guild_id = guild_id
        return new
    
    
    @property
    def guild(self):
        """
        Returns the guild enhancement entitlements create / delete event's guild.
        
        Returns
        -------
        guild : ``None | Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    @property
    def entitlement(self):
        """
        Returns the first entitlement of the event.
        
        Returns
        -------
        entitlement : ``None | Entitlement``
        """
        entitlements = self.entitlements
        if (entitlements is not None):
            return entitlements[0]
    
    
    def iter_entitlements(self):
        """
        Iterates over the entitlements of the event.
        
        This method is an iterable generator.
        
        Yields
        ------
        entitlement : ``Entitlement``
        """
        entitlements = self.entitlements
        if (entitlements is not None):
            yield from entitlements



__all__ = ('EntitySelectDefaultValue',)

from scarletio import RichAttributeErrorBaseType

from .fields import parse_id, parse_type, put_id, put_type, validate_id, validate_type


class EntitySelectDefaultValue(RichAttributeErrorBaseType):
    """
    Entity presented in an entity select by default.
    
    Attributes
    ----------
    id : `int`
        The represented entity's identifier.
    
    type : ``EntitySelectDefaultValueType``
        The represented entity's type.
    """
    def __new__(cls, option_type, entity_id):
        """
        Creates a new component option with the given parameters.
        
        Parameters
        ----------
        option_type : ``str | EntitySelectDefaultValueType``
            The represented entity's type.
        
        entity_id : `int | str`
            The represented entity's identifier.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        entity_id = validate_id(entity_id)
        option_type = validate_type(option_type)
        
        # Construct
        self = object.__new__(cls)
        self.id = entity_id
        self.type = option_type
        return self
    
    
    @classmethod
    def from_fields(cls, option_type, entity_id):
        """
        Creates a new entity select default option with the given fields.
        
        > This constructor does not validation, ment for internal use.
        
        Parameters
        ----------
        option_type : ``EntitySelectDefaultValueType``
            The represented entity's type.
        
        entity_id : `int`
            The represented entity's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        # Construct
        self = object.__new__(cls)
        self.id = entity_id
        self.type = option_type
        return self
    

    @classmethod
    def from_data(cls, data):
        """
        Creates a new string select option from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            String select option data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.id = parse_id(data)
        self.type = parse_type(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the string select option to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        
        put_id(self.id, data, defaults)
        put_type(self.type, data, defaults)
        
        return data
    
    
    def __repr__(self):
        """Returns the string select option's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # id
        repr_parts.append(', id = ')
        repr_parts.append(repr(self.id))
        
         # type
        option_type = self.type
        repr_parts.append(', type = ')
        repr_parts.append(option_type.name)
        repr_parts.append(' ~ ')
        repr_parts.append(repr(option_type.value))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two string select options are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # id
        if self.id != other.id:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the string select option's hash value."""
        hash_value = 0
        
        # id
        hash_value ^= self.id
        
        # type
        hash_value ^= hash(self.type)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the string select option.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.id = self.id
        new.type = self.type
        return new
    
    
    def copy_with(self, entity_id = ..., option_type = ...):
        """
        Copes the string select with modifying it's defined attributes.
        
        Parameters
        ----------
        entity_id : `int | str`, Optional (Keyword only)
            The represented entity's identifier.
        
        option_type : ``str | EntitySelectDefaultValueType``, Optional (Keyword only)
            The represented entity's type.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # entity_id
        if entity_id is ...:
            entity_id = self.id
        else:
            entity_id = validate_id(entity_id)
        
        # option_type
        if option_type is ...:
            option_type = self.type
        else:
            option_type = validate_type(option_type)
        
        # Construct
        new = object.__new__(type(self))
        new.id = entity_id
        new.type = option_type
        return new



__all__ = ()

from ...field_parsers import bool_parser_factory, entity_id_parser_factory, nullable_date_time_parser_factory
from ...field_putters import (
    bool_optional_putter_factory, entity_id_putter_factory, nullable_date_time_optional_putter_factory
)
from ...field_validators import bool_validator_factory, entity_id_validator_factory, nullable_date_time_validator_factory
from ...user import ClientUserBase

from ..guild import Guild


# ended

parse_ended = bool_parser_factory('ended', False)
put_ended = bool_optional_putter_factory('ended', False)
validate_ended = bool_validator_factory('ended', False)


# ends_at

parse_ends_at = nullable_date_time_parser_factory('ends_at')
put_ends_at = nullable_date_time_optional_putter_factory('ends_at')
validate_ends_at = nullable_date_time_validator_factory('ends_at')


# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id = entity_id_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', Guild)


# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('guild_boost_id')


# paused_until

parse_paused_until = nullable_date_time_parser_factory('pause_ends_at')
put_paused_until = nullable_date_time_optional_putter_factory('pause_ends_at')
validate_paused_until = nullable_date_time_validator_factory('paused_until')


# user_id

parse_user_id = entity_id_parser_factory('user_id')
put_user_id = entity_id_putter_factory('user_id')
validate_user_id = entity_id_validator_factory('user_id', ClientUserBase)



__all__ = ()

from ...field_parsers import nullable_string_parser_factory
from ...field_putters import nullable_string_optional_putter_factory
from ...field_validators import nullable_string_validator_factory

# image_large

parse_image_large = nullable_string_parser_factory('large_image')
put_image_large = nullable_string_optional_putter_factory('large_image')
validate_image_large = nullable_string_validator_factory('image_large', 0, 1024)


# image_small

parse_image_small = nullable_string_parser_factory('small_image')
put_image_small = nullable_string_optional_putter_factory('small_image')
validate_image_small = nullable_string_validator_factory('image_small', 0, 1024)


# text_large

parse_text_large = nullable_string_parser_factory('large_text')
put_text_large = nullable_string_optional_putter_factory('large_text')
validate_text_large = nullable_string_validator_factory('text_large', 0, 1024)


# url_large

parse_url_large = nullable_string_parser_factory('large_url')
put_url_large = nullable_string_optional_putter_factory('large_url')
validate_url_large = nullable_string_validator_factory('url_large', 0, 1024)


# text_small

parse_text_small = nullable_string_parser_factory('small_text')
put_text_small = nullable_string_optional_putter_factory('small_text')
validate_text_small = nullable_string_validator_factory('text_small', 0, 1024)


# url_small

parse_url_small = nullable_string_parser_factory('small_url')
put_url_small = nullable_string_optional_putter_factory('small_url')
validate_url_small = nullable_string_validator_factory('url_small', 0, 1024)


# image_invite_cover

parse_image_invite_cover = nullable_string_parser_factory('invite_cover_image')
put_image_invite_cover = nullable_string_optional_putter_factory('invite_cover_image')
validate_image_invite_cover = nullable_string_validator_factory('image_invite_cover', 0, 1024)



__all__ = ()

from ...application_command.application_command_permission.fields import validate_permission_overwrites

from ..audit_log_entry_change_conversion import AuditLogEntryChangeConversion, AuditLogEntryChangeConversionGroup
from ..audit_log_entry_change_conversion.change_deserializers import (
    change_deserializer_application_command_permission_overwrite
)
from ..audit_log_entry_change_conversion.change_serializers import (
    change_serializer_application_command_permission_overwrite
)
from ..audit_log_entry_change_conversion.key_pre_checks import key_pre_check_id
from ..audit_log_entry_change_conversion.value_mergers import value_merger_sorted_array


# ---- permission_overwrite ----

PERMISSION_OVERWRITE_CONVERSION = AuditLogEntryChangeConversion(
    ('\\d+',),
    'permission_overwrites',
    change_deserialization_key_pre_check = key_pre_check_id,
    change_deserializer = change_deserializer_application_command_permission_overwrite,
    change_serializer = change_serializer_application_command_permission_overwrite,
    value_merger = value_merger_sorted_array,
    value_validator = validate_permission_overwrites
)

# ---- Construct ----

APPLICATION_COMMAND_CONVERSIONS = AuditLogEntryChangeConversionGroup(
    PERMISSION_OVERWRITE_CONVERSION,
)



__all__ = ()

from functools import partial as partial_func

from .fields import (
    put_allow, put_deny, put_target_id, put_target, put_target_type, validate_allow,
    validate_deny, validate_target, validate_target_id, validate_target_type
)


PERMISSION_OVERWRITE_PERMISSION_FIELD_CONVERTERS = {
    'allow': (validate_allow, put_allow),
    'deny': (validate_deny, put_deny),
}


PERMISSION_OVERWRITE_FIELD_CONVERTERS = {
    **PERMISSION_OVERWRITE_PERMISSION_FIELD_CONVERTERS,
    'target': (validate_target, partial_func(put_target, include_internals = True)),
    'target_id': (validate_target_id, put_target_id),
    'target_type': (validate_target_type, put_target_type),
}



__all__ = ()

from ...field_parsers import (
    entity_id_parser_factory, force_string_parser_factory, int_parser_factory, nullable_string_parser_factory
)
from ...field_putters import (
    entity_id_putter_factory, int_putter_factory, nullable_string_putter_factory, url_optional_putter_factory
)
from ...field_validators import (
    entity_id_validator_factory, force_string_validator_factory, int_conditional_validator_factory,
    nullable_string_validator_factory, url_optional_validator_factory
)

from ...message import Attachment

# attachment_id

parse_attachment_id = entity_id_parser_factory('attachment_id')
put_attachment_id = entity_id_putter_factory('attachment_id')
validate_attachment_id = entity_id_validator_factory('attachment_id', Attachment)


# content_type

parse_content_type = nullable_string_parser_factory('content_type')
put_content_type = nullable_string_putter_factory('content_type')
validate_content_type = nullable_string_validator_factory('content_type', 0, 1024)


# height

parse_height = int_parser_factory('height', 0)
put_height = int_putter_factory('height')
validate_height = int_conditional_validator_factory(
    'height',
    0,
    lambda height : height >= 0,
    '>= 0',
)


# proxy_url

parse_proxy_url = nullable_string_parser_factory('proxy_url')
put_proxy_url = url_optional_putter_factory('proxy_url')
validate_proxy_url = url_optional_validator_factory('proxy_url')


# url

parse_url = force_string_parser_factory('url')
put_url = url_optional_putter_factory('url')
validate_url = force_string_validator_factory('url', 0, 1024)


# width

parse_width = int_parser_factory('width', 0)
put_width = int_putter_factory('width')
validate_width = int_conditional_validator_factory(
    'width',
    0,
    lambda width : width >= 0,
    '>= 0',
)



__all__ = ('LabelType', 'ReleasePhase',)

from ...bases import Preinstance as P, PreinstancedBase


class LabelType(PreinstancedBase, value_type = int):
    """
    Represents a embedded activity's label's type.
    
    Attributes
    ----------
    name : `str`
        The name of the label's type.
    
    value : `int`
        The identifier value the label's type.
    
    Type Attributes
    ---------------
    Every predefined label's type can be accessed as type attribute as well:
    
    +-----------------------+-----------+-------+
    | Type attribute name   | Name      | Value |
    +=======================+===========+=======+
    | none                  | none      | 0     |
    +-----------------------+-----------+-------+
    | new                   | new       | 1     |
    +-----------------------+-----------+-------+
    | updated               | updated   | 2     |
    +-----------------------+-----------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    DEFAULT_NAME = 'UNDEFINED'
    
    __slots__ = ()
    
    none = P(0, 'none')
    new = P(1, 'new')
    updated = P(2, 'updated')


class ReleasePhase(PreinstancedBase, value_type = str):
    """
    Represents the release phase of an embedded activity on a platform.
    
    Attributes
    ----------
    name : `str`
        The name of the release phase.
    
    value : `str`
        The Discord side identifier value of the platform.
    
    Type Attributes
    ---------------
    Every predefined release phase can also be accessed as type attribute:
    
    +-----------------------+---------------+---------------+
    | Type attribute name   | name          | value         |
    +=======================+===============+===============+
    | global_launch         | global launch | global_launch |
    +-----------------------+---------------+---------------+
    """
    __slots__ = ()

    # predefined
    global_launch = P('global_launch', 'global launch')


ReleasePhase.INSTANCES[''] = ReleasePhase.global_launch



from .emoji import *
from .parsing import *
from .reaction import *
from .reaction_events import *
from .reaction_mapping import *
from .reaction_mapping_line import *
from .unicode import *


__all__ = (
    *emoji.__all__,
    *parsing.__all__,
    *reaction.__all__,
    *reaction_events.__all__,
    *reaction_mapping.__all__,
    *reaction_mapping_line.__all__,
    *unicode.__all__,
)



from .poll import *
from .poll_answer import *
from .poll_events import *
from .poll_question import *
from .poll_result import *


__all__ = (
    *poll.__all__,
    *poll_answer.__all__,
    *poll_events.__all__,
    *poll_question.__all__,
    *poll_result.__all__,
)



__all__ = ()

from .fields import (
    put_privacy_level, put_scheduled_event_id, put_send_start_notification, put_topic,
    validate_privacy_level, validate_scheduled_event_id, validate_send_start_notification, validate_topic
)


STAGE_CREATE_FIELD_CONVERTERS = {
    'privacy_level': (validate_privacy_level, put_privacy_level),
    'scheduled_event': (validate_scheduled_event_id, put_scheduled_event_id),
    'scheduled_event_id': (validate_scheduled_event_id, put_scheduled_event_id),
    'send_start_notification': (validate_send_start_notification, put_send_start_notification),
    'topic': (validate_topic, put_topic),
}

STAGE_EDIT_FIELD_CONVERTERS = {
    'privacy_level': (validate_privacy_level, put_privacy_level),
    'topic': (validate_topic, put_topic),
}



__all__ = ()

from ...color import Color
from ...field_parsers import (
    entity_id_array_parser_factory, entity_id_parser_factory, force_string_parser_factory, int_parser_factory,
    nullable_string_parser_factory, preinstanced_array_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    entity_id_putter_factory, force_string_putter_factory, int_optional_putter_factory, int_putter_factory,
    nullable_string_putter_factory, optional_entity_id_array_optional_putter_factory, preinstanced_array_putter_factory,
    preinstanced_putter_factory
)
from ...field_validators import (
    entity_id_array_validator_factory, entity_id_validator_factory, force_string_validator_factory,
    int_conditional_validator_factory, int_options_validator_factory, nullable_object_array_validator_factory,
    nullable_string_validator_factory, preinstanced_array_validator_factory, preinstanced_validator_factory
)
from ...scheduled_event import PrivacyLevel

from ..guild import GuildFeature
from ..guild.guild_boost_perks import LEVELS
from ..guild_activity_overview_activity import GuildActivityOverviewActivity
from ..guild_activity_overview_tag import GuildActivityOverviewTag

from .constants import (
    BADGE_TAG_LENGTH_MAX, BADGE_TAG_LENGTH_MIN, DESCRIPTION_LENGTH_MAX, NAME_LENGTH_MAX, NAME_LENGTH_MIN
)


# activities

def parse_activities(data):
    """
    Parses the activities from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    activities : `None | dict<int, GuildActivityOverviewActivity>`
    """
    activities_datas = data.get('game_activity', None)
    if (activities_datas is not None) and activities_datas:
        return {
            int(application_id): GuildActivityOverviewActivity.from_data(activity_data)
            for application_id, activity_data in activities_datas.items()
        }


def put_activities(activities, data, defaults):
    """
    Puts activity into the given data.
    
    Parameters
    ----------
    activities : `None | dict<int, GuildActivityOverviewActivity>`
        Activities to serialize.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if activities is None:
        activities_data = {}
    else:
        activities_data = {
            str(application_id): guild_activity_overview_activity.to_data(defaults = defaults)
            for application_id, guild_activity_overview_activity in activities.items()
        }
    
    data['game_activity'] = activities_data
    return data


def validate_activities(activities):
    """
    Validates the given activities.
    
    Parameters
    ----------
    activities : `object`
        Activities to validate.
    
    Returns
    -------
    activities : `None | dict<int, GuildActivityOverviewActivity>`
    
    Raises
    ------
    TypeError
        - If `activities` type is incorrect.
    """
    if activities is None:
        return None
    
    if not isinstance(activities, dict):
        raise TypeError(
            f'`activities` can be `None` or `dict`, got {type(activities).__name__}; {activities!r}.'
        )
    
    validated_activities = None
    
    for application_id, guild_activity_overview_activity in activities.items():
        if validated_activities is None:
            validated_activities = {}
        
        if not isinstance(application_id, int):
            raise TypeError(
                f'`activities` keys can be `int`, got {type(application_id).__name__}; {application_id!r}.'
            )
        
        if not isinstance(guild_activity_overview_activity, GuildActivityOverviewActivity):
            raise TypeError(
                f'`activities` keys can be `{GuildActivityOverviewActivity.__name__}`, '
                f'got {type(guild_activity_overview_activity).__name__}; {guild_activity_overview_activity!r}.'
            )
        
        validated_activities[application_id] = guild_activity_overview_activity
        
    return validated_activities


# activity_application_ids

parse_activity_application_ids = entity_id_array_parser_factory('game_application_ids', ordered = False)
put_activity_application_ids = optional_entity_id_array_optional_putter_factory('game_application_ids')
validate_activity_application_ids = entity_id_array_validator_factory(
    'activity_application_ids', NotImplemented, include = 'Application', ordered = False
)


# approximate_online_count

parse_approximate_online_count = int_parser_factory('online_count', 0)
put_approximate_online_count = int_putter_factory('online_count')
validate_approximate_online_count = int_conditional_validator_factory(
    'approximate_online_count',
    0,
    lambda approximate_online_count : approximate_online_count >= 0,
    '>= 0',
)


# approximate_user_count

parse_approximate_user_count = int_parser_factory('member_count', 0)
put_approximate_user_count = int_putter_factory('member_count')
validate_approximate_user_count = int_conditional_validator_factory(
    'approximate_user_count',
    0,
    lambda approximate_user_count : approximate_user_count >= 0,
    '>= 0',
)


# badge

parse_badge = int_parser_factory('badge', 0)
put_badge = int_putter_factory('badge')
validate_badge = int_conditional_validator_factory(
    'badge',
    0,
    lambda badge : badge >= 0,
    '>= 0',
)


# badge_color_primary

def parse_badge_color_primary(data):
    """
    Parses banner color (custom) from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        User data.
    
    Returns
    -------
    badge_color_primary : `None | Color`
    """
    badge_color_primary = data.get('badge_color_primary', None)
    if (badge_color_primary is not None):
        badge_color_primary = Color.from_html(badge_color_primary)
    
    return badge_color_primary


def put_badge_color_primary(badge_color_primary, data, defaults):
    """
    Serializes the given color into the given data.
    
    Parameters
    ----------
    badge_color_primary : `None | Color`
        Banner color.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (badge_color_primary is not None):
        badge_color_primary = badge_color_primary.as_html
    data['badge_color_primary'] = badge_color_primary
    
    return data


def validate_badge_color_primary(badge_color_primary):
    """
    Validates the given `badge_color_primary` value.
    
    Parameters
    ----------
    badge_color_primary : `None | Color | value`
        The banner color to validate.
    
    Returns
    -------
    badge_color_primary : `None | Color`
    
    Raises
    ------
    TypeError
        - If `badge_color_primary` is not `None`, `Color`, `int`.
    ValueError
        - If `badge_color_primary`'s value is out of the expected range.
    """
    if (badge_color_primary is None):
        return None
    
    if isinstance(badge_color_primary, Color):
        return badge_color_primary
    
    if isinstance(badge_color_primary, int):
        if badge_color_primary < 0 or badge_color_primary > 0xffffff:
            raise ValueError(
                f'`badge_color_primary` can be between 0 and 0xffffff, got {badge_color_primary!r}.'
            )
        
        return Color(badge_color_primary)
    
    raise TypeError(
        f'`badge_color_primary` can be `None`, `{Color.__name__}`, `int`, '
        f'got {type(badge_color_primary).__name__}; {badge_color_primary!r}.'
    )


# badge_color_secondary

def parse_badge_color_secondary(data):
    """
    Parses banner color (custom) from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        User data.
    
    Returns
    -------
    badge_color_secondary : `None | Color`
    """
    badge_color_secondary = data.get('badge_color_secondary', None)
    if (badge_color_secondary is not None):
        badge_color_secondary = Color.from_html(badge_color_secondary)
    
    return badge_color_secondary


def put_badge_color_secondary(badge_color_secondary, data, defaults):
    """
    Serializes the given color into the given data.
    
    Parameters
    ----------
    badge_color_secondary : `None | Color`
        Banner color.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (badge_color_secondary is not None):
        badge_color_secondary = badge_color_secondary.as_html
    data['badge_color_secondary'] = badge_color_secondary
    
    return data


def validate_badge_color_secondary(badge_color_secondary):
    """
    Validates the given `badge_color_secondary` value.
    
    Parameters
    ----------
    badge_color_secondary : `None | Color | value`
        The banner color to validate.
    
    Returns
    -------
    badge_color_secondary : `None | Color`
    
    Raises
    ------
    TypeError
        - If `badge_color_secondary` is not `None`, `Color`, `int`.
    ValueError
        - If `badge_color_secondary`'s value is out of the expected range.
    """
    if (badge_color_secondary is None):
        return None
    
    if isinstance(badge_color_secondary, Color):
        return badge_color_secondary
    
    if isinstance(badge_color_secondary, int):
        if badge_color_secondary < 0 or badge_color_secondary > 0xffffff:
            raise ValueError(
                f'`badge_color_secondary` can be between 0 and 0xffffff, got {badge_color_secondary!r}.'
            )
        
        return Color(badge_color_secondary)
    
    raise TypeError(
        f'`badge_color_secondary` can be `None`, `{Color.__name__}`, `int`, '
        f'got {type(badge_color_secondary).__name__}; {badge_color_secondary!r}.'
    )


# badge_tag

parse_badge_tag = force_string_parser_factory('tag')
put_badge_tag = force_string_putter_factory('tag')
validate_badge_tag = force_string_validator_factory('badge_tag', BADGE_TAG_LENGTH_MIN, BADGE_TAG_LENGTH_MAX)


# banner_color

def parse_banner_color(data):
    """
    Parses banner color from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        User data.
    
    Returns
    -------
    banner_color : `None | Color`
    """
    banner_color = data.get('brand_color_primary', None)
    if (banner_color is not None):
        banner_color = Color.from_html(banner_color)
    
    return banner_color


def put_banner_color(banner_color, data, defaults):
    """
    Serializes the given color into the given data.
    
    Parameters
    ----------
    banner_color : `None | Color`
        Banner color.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (banner_color is not None):
        if (banner_color is not None):
            banner_color = banner_color.as_html
        data['brand_color_primary'] = banner_color
    
    return data


def validate_banner_color(banner_color):
    """
    Validates the given `banner_color` value.
    
    Parameters
    ----------
    banner_color : `None | Color | value`
        The banner color to validate.
    
    Returns
    -------
    banner_color : `None | Color`
    
    Raises
    ------
    TypeError
        - If `banner_color` is not `None`, `Color`, `int`.
    ValueError
        - If `banner_color`'s value is out of the expected range.
    """
    if (banner_color is None):
        return None
    
    if isinstance(banner_color, Color):
        return banner_color
    
    if isinstance(banner_color, int):
        if banner_color < 0 or banner_color > 0xffffff:
            raise ValueError(
                f'`banner_color` can be between 0 and 0xffffff, got {banner_color!r}.'
            )
        
        return Color(banner_color)
    
    raise TypeError(
        f'`banner_color` can be `None`, `{Color.__name__}`, `int`, '
        f'got {type(banner_color).__name__}; {banner_color!r}.'
    )


# boost_count

parse_boost_count = int_parser_factory('premium_subscription_count', 0)
put_boost_count = int_optional_putter_factory('premium_subscription_count', 0)
validate_boost_count = int_conditional_validator_factory(
    'boost_count',
    0,
    lambda boost_count : boost_count >= 0,
    '>= 0',
)


# boost_level

parse_boost_level = int_parser_factory('premium_tier', 0)
put_boost_level = int_putter_factory('premium_tier')
validate_boost_level = int_options_validator_factory('boost_level', frozenset(LEVELS.keys()), 0)


# description

parse_description = nullable_string_parser_factory('description')
put_description = nullable_string_putter_factory('description')
validate_description = nullable_string_validator_factory('description', 0, DESCRIPTION_LENGTH_MAX)


# features

parse_features = preinstanced_array_parser_factory('features', GuildFeature)
put_features = preinstanced_array_putter_factory('features')
validate_features = preinstanced_array_validator_factory('features', GuildFeature)


# id

parse_id = entity_id_parser_factory('id')
put_id = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('guild_id')


# name

parse_name = force_string_parser_factory('name')
put_name = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)


# privacy_level

parse_privacy_level = preinstanced_parser_factory('visibility', PrivacyLevel, PrivacyLevel.guild_only)
put_privacy_level = preinstanced_putter_factory('visibility')
validate_privacy_level = preinstanced_validator_factory('privacy_level', PrivacyLevel)


# tags

def _tag_data_sort_key_getter(tag_data):
    """
    Sort key getter used to sort tag datas based on their position.
    
    Parameters
    ----------
    tag_data : `dict<str, object>`
        Tag data.
    
    Returns
    -------
    key : `int`
    """
    return tag_data['position']


def parse_tags(data):
    """
    Parses the tags from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    tags : `None | tuple<GuildActivityOverviewTag>`
    """
    tag_datas = data.get('traits', None)
    if (tag_datas is not None) and tag_datas:
        return (*(
            GuildActivityOverviewTag.from_data(tag_data)
            for tag_data in sorted(tag_datas, key = _tag_data_sort_key_getter)
        ),)


def put_tags(tags, data, defaults):
    """
    Puts tag into the given data.
    
    Parameters
    ----------
    tags : `None | tuple<GuildActivityOverviewTag>`
        Activities to serialize.
    
    data : `dict<str, object>`
        Json serializable dictionary.
    
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    tag_datas = []
    
    if (tags is not None):
        for index, guild_tag_overview_tag in enumerate(tags):
            tag_data = guild_tag_overview_tag.to_data(defaults = defaults)
            tag_data['position'] = index
            tag_datas.append(tag_data)
    
    data['traits'] = tag_datas
    return data


validate_tags = nullable_object_array_validator_factory('tags', GuildActivityOverviewTag)



__all__ = ()

BOT_PUBLIC_DEFAULT = False
BOT_REQUIRES_CODE_GRANT_DEFAULT = False
OVERLAY_DEFAULT = False
OVERLAY_COMPATIBILITY_HOOK_DEFAULT = False
HOOK_DEFAULT = False
INTEGRATION_PUBLIC_DEFAULT = False
INTEGRATION_REQUIRES_CODE_GRANT_DEFAULT = False


MAX_PARTICIPANTS_DEFAULT = 0

NAME_LENGTH_MIN = 2
NAME_LENGTH_MAX = 128

DESCRIPTION_LENGTH_MAX = 400


# Experimental addition

EMBEDDED_ACTIVITY_NAME_TO_APPLICATION_ID = {
    'Betrayal.io': 773336526917861400,
    'Chess in The Park': 832012774040141894,
    'Fishington.io': 814288819477020702,
    'Poker Night': 755827207812677713,
    'Land-io': 903769130790969345,
    'Sketch Heads': 902271654783242291,
    'Letter League': 879863686565621790,
    'Checkers In The Park': 832013003968348200,
    'Chef Showdown': 1037680572660727838,
    'Bobble League': 947957217959759964,
    'Krunker Strike FRVR': 1011683823555199066,
    'Gartic Phone': 1007373802981822582,
    'Whiteboard': 1070087967294631976,
    'Know What I Meme': 1078728822972764312,
    'Color Together': 1039835161136746497,
    'Bobble Bash': 1107689944685748377,
    
    # Spell Cast old name.
    # 'Spell Cast': 852509694341283871,
    'SpellCast': 852509694341283871,
    
    # Ocho old name.
    # 'Ocho': 832025144389533716,
    'Blazing 8s': 832025144389533716,
    
    # Application id changed.
    # 'Putt Party': 910224161476083792,
    'Putt Party': 945737671223947305,
    
    # Many versions uoh.
    # 'YouTube Together': 755600276941176913,
    'Watch Together': 880218394199220334,
    # 'Watch Together Dev': 880218832743055411,
}


EMBEDDED_ACTIVITY_APPLICATION_ID_TO_NAME = {
    value: key for key, value in EMBEDDED_ACTIVITY_NAME_TO_APPLICATION_ID.items()
}



__all__ = ('IntegrationExpireBehavior',)

from ...bases import Preinstance as P, PreinstancedBase


class IntegrationExpireBehavior(PreinstancedBase, value_type = int):
    """
    Represents an ``IntegrationDetail``'s expire behavior.
    
    Attributes
    ----------
    name : `str`
        The name of the integration expire behavior.
    
    value : `int`
        The Discord side identifier value of the integration expire behavior.
    
    Type Attributes
    ---------------
    Every predefined expire behavior can be accessed as type attribute as well:
    
    +-----------------------+---------------+-------+
    | Type attribute name   | name          | value |
    +=======================+===============+=======+
    | remove_role           | remove_role   | 0     |
    +-----------------------+---------------+-------+
    | kick                  | kick          | 1     |
    +-----------------------+---------------+-------+
    """
    __slots__ = ()
    
    # predefined
    remove_role = P(0, 'remove_role')
    kick = P(1, 'kick')



__all__ = ('VoiceState', )

from scarletio import RichAttributeErrorBaseType, include

from ...core import GUILDS

from ..user import create_partial_user_from_id

from .fields import (
    parse_channel_id, parse_deaf, parse_mute, parse_requested_to_speak_at, parse_self_deaf, parse_self_mute,
    parse_self_stream, parse_self_video, parse_session_id, parse_speaker, parse_user_id, put_channel_id,
    put_deaf, put_mute, put_requested_to_speak_at, put_self_deaf, put_self_mute,
    put_self_stream, put_self_video, put_session_id, put_speaker, put_user_id,
    validate_channel_id, validate_deaf, validate_guild_id, validate_mute, validate_requested_to_speak_at,
    validate_self_deaf, validate_self_mute, validate_self_stream, validate_self_video, validate_session_id,
    validate_speaker, validate_user_id
)


ChannelType = include('ChannelType')
create_partial_channel_from_id = include('create_partial_channel_from_id')
create_partial_role_from_id = include('create_partial_role_from_id')


class VoiceState(RichAttributeErrorBaseType):
    """
    Represents a user at a ``Channel``.
    
    Attributes
    ----------
    _cache_user : ``ClientUserBase``
        The voice state's respective user.
    channel_id : `int`
        The channel's identifier to where the user is connected to.
    deaf : `bool`
        Whether the user is deafen.
    guild_id : `int`
        The guild's identifier where the user is connected to.
    mute : `bool`
        Whether the user is muted.
    requested_to_speak_at : `None | DateTime`
        When the user requested to speak.
        
        Only applicable in stage channels.
    
    self_deaf : `bool`
        Whether the user muted everyone else.
    self_mute : `bool`
        Whether the user muted itself.
    self_stream : `bool`
        Whether the user screen shares with the go live option.
    self_video : `bool`
        Whether the user sends video from a camera source.
    session_id : `str``
        The user's voice session id.
    speaker : `bool`
        Whether the user is suppressed inside of the voice channel.
        
        At voice channels it is always `False`, but in stage channels it can vary.
    
    user_id : `int`
        The voice state's respective user's identifier.
    """
    __slots__ = (
        '_cache_user', 'channel_id', 'deaf', 'guild_id', 'mute', 'requested_to_speak_at', 'self_deaf', 'self_mute',
        'self_stream', 'self_video', 'session_id', 'speaker', 'user_id'
    )
    
    
    def __new__(
        cls,
        *,
        channel_id = ...,
        deaf = ...,
        guild_id = ...,
        speaker = ...,
        mute = ...,
        requested_to_speak_at = ...,
        self_deaf = ...,
        self_mute = ...,
        self_stream = ...,
        self_video = ...,
        session_id = ...,
        user_id = ...,
    ):
        """
        Creates a new voice state with the given fields.
        
        Parameters
        ----------
        channel_id : `int`, Optional (Keyword only)
            The channel's identifier to where the user is connected to.
        deaf : `bool`, Optional (Keyword only)
            Whether the user is deafen.
        guild_id : `int`, Optional (Keyword only)
            The guild's identifier where the user is connected to.
        mute : `bool`, Optional (Keyword only)
            Whether the user is muted.
        requested_to_speak_at : `None | DateTime`, Optional (Keyword only)
            When the user requested to speak.
        self_deaf : `bool`, Optional (Keyword only)
            Whether the user muted everyone else.
        self_mute : `bool`, Optional (Keyword only)
            Whether the user muted itself.
        self_stream : `bool`, Optional (Keyword only)
            Whether the user screen shares with the go live option.
        self_video : `bool`, Optional (Keyword only)
            Whether the user sends video from a camera source.
        session_id : `str``, Optional (Keyword only)
            The user's voice session id.
        speaker : `bool`, Optional (Keyword only)
            Whether the user is suppressed inside of the voice channel.
        user_id : `int`, Optional (Keyword only)
            The voice state's respective user's identifier.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # channel_id
        if channel_id is ...:
            channel_id = 0
        else:
            channel_id = validate_channel_id(channel_id)
        
        # deaf
        if deaf is ...:
            deaf = False
        else:
            deaf = validate_deaf(deaf)
        
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_guild_id(guild_id)
        
        # mute
        if mute is ...:
            mute = False
        else:
            mute = validate_mute(mute)
        
        # requested_to_speak_at
        if requested_to_speak_at is ...:
            requested_to_speak_at = None
        else:
            requested_to_speak_at = validate_requested_to_speak_at(requested_to_speak_at)
        
        # self_deaf
        if self_deaf is ...:
            self_deaf = False
        else:
            self_deaf = validate_self_deaf(self_deaf)
        
        # self_mute
        if self_mute is ...:
            self_mute = False
        else:
            self_mute = validate_self_mute(self_mute)
        
        # self_stream
        if self_stream is ...:
            self_stream = False
        else:
            self_stream = validate_self_stream(self_stream)
        
        # self_video
        if self_video is ...:
            self_video = False
        else:
            self_video = validate_self_video(self_video)
        
        # session_id
        if session_id is ...:
            session_id = ''
        else:
            session_id = validate_session_id(session_id)
        
        # speaker
        if speaker is ...:
            speaker = False
        else:
            speaker = validate_speaker(speaker)
        
        # user_id
        if user_id is ...:
            user_id = 0
        else:
            user_id = validate_user_id(user_id)
        
        # Construct
        self = object.__new__(cls)
        self._cache_user = None
        self.channel_id = channel_id
        self.deaf = deaf
        self.guild_id = guild_id
        self.mute = mute
        self.requested_to_speak_at = requested_to_speak_at
        self.self_deaf = self_deaf
        self.self_mute = self_mute
        self.self_stream = self_stream
        self.self_video = self_video
        self.session_id = session_id
        self.speaker = speaker
        self.user_id = user_id
        return self
    
    
    @classmethod
    def from_data(cls, data, guild_id, *, strong_cache = True):
        """
        Creates a voice state object from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Voice state data.
        guild_id : `int`
            The voice state's guild's identifier.
        strong_cache : `bool` = `True`, Optional (Keyword only)
            Whether the instance should be put into its strong cache.
        
        Returns
        -------
        new : `None`, `instance<cls>`
        """
        channel_id = parse_channel_id(data)
        if not channel_id:
            return
        
        user_id = parse_user_id(data)
        
        if strong_cache:
            guild = GUILDS.get(guild_id, None)
            if (guild is not None):
                voice_states = guild.voice_states
                if (voice_states is not None):
                    try:
                        return voice_states[user_id]
                    except KeyError:
                        pass
        
        self = object.__new__(cls)
        self._cache_user = None
        self.channel_id = channel_id
        self.deaf = parse_deaf(data)
        self.guild_id = guild_id
        self.speaker = parse_speaker(data)
        self.mute = parse_mute(data)
        self.requested_to_speak_at = parse_requested_to_speak_at(data)
        self.self_deaf = parse_self_deaf(data)
        self.self_mute = parse_self_mute(data)
        self.self_stream = parse_self_stream(data)
        self.self_video = parse_self_video(data)
        self.session_id = parse_session_id(data)
        self.user_id = user_id
        
        if strong_cache:
            if (guild is not None):
                if voice_states is None:
                    voice_states = {}
                    guild.voice_states = voice_states
                
                voice_states[user_id] = self
        
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the voice state to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with the default values should be included as well.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_channel_id(self.channel_id, data, defaults)
        put_deaf(self.deaf, data, defaults)
        put_mute(self.mute, data, defaults)
        put_requested_to_speak_at(self.requested_to_speak_at, data, defaults)
        put_self_deaf(self.self_deaf, data, defaults)
        put_self_mute(self.self_mute, data, defaults)
        put_self_stream(self.self_stream, data, defaults)
        put_self_video(self.self_video, data, defaults)
        put_session_id(self.session_id, data, defaults)
        put_speaker(self.speaker, data, defaults)
        put_user_id(self.user_id, data, defaults)
        return data
    
    
    def _update_attributes(self, data):
        """
        Updates the voice state with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Voice state data.
        """
        self.deaf = parse_deaf(data)
        self.mute = parse_mute(data)
        self.requested_to_speak_at = parse_requested_to_speak_at(data)
        self.self_deaf = parse_self_deaf(data)
        self.self_mute = parse_self_mute(data)
        self.self_stream = parse_self_stream(data)
        self.self_video = parse_self_video(data)
        self.speaker = parse_speaker(data)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the voice state and returns it's overwritten attributes as a `dict` with a `attribute-name` -
        `old-value` relation.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Voice state data.
        
        Returns
        -------
        old_attributes : `dict<str, object>`
            All item in the returned dictionary is optional.
        
        Returned Data Structure
        -----------------------
        
        +-----------------------+-----------------------+
        | Keys                  | Values                |
        +=======================+=======================+
        | deaf                  | `str`                 |
        +-----------------------+-----------------------+
        | mute                  | `bool`                |
        +-----------------------+-----------------------+
        | requested_to_speak_at | `None | DateTime`     |
        +-----------------------+-----------------------+
        | self_deaf             | `bool`                |
        +-----------------------+-----------------------+
        | self_mute             | `bool`                |
        +-----------------------+-----------------------+
        | self_stream           | `bool`                |
        +-----------------------+-----------------------+
        | self_video            | `bool`                |
        +-----------------------+-----------------------+
        | speaker               | `bool`                |
        +-----------------------+-----------------------+
        """
        old_attributes = {}
        
        deaf = parse_deaf(data)
        if self.deaf != deaf:
            old_attributes['deaf'] = self.deaf
            self.deaf = deaf
        
        mute = parse_mute(data)
        if self.mute != mute:
            old_attributes['mute'] = self.mute
            self.mute = mute
        
        requested_to_speak_at = parse_requested_to_speak_at(data)
        if self.requested_to_speak_at != requested_to_speak_at:
            old_attributes['requested_to_speak_at'] = self.requested_to_speak_at
            self.requested_to_speak_at = requested_to_speak_at
        
        self_deaf = parse_self_deaf(data)
        if self.self_deaf != self_deaf:
            old_attributes['self_deaf'] = self.self_deaf
            self.self_deaf = self_deaf
        
        self_video = parse_self_video(data)
        if self.self_video != self_video:
            old_attributes['self_video'] = self.self_video
            self.self_video = self_video
        
        self_stream = parse_self_stream(data)
        if self.self_stream != self_stream:
            old_attributes['self_stream'] = self.self_stream
            self.self_stream = self_stream
        
        self_mute = parse_self_mute(data)
        if self.self_mute != self_mute:
            old_attributes['self_mute'] = self.self_mute
            self.self_mute = self_mute
        
        speaker = parse_speaker(data)
        if self.speaker != speaker:
            old_attributes['speaker'] = self.speaker
            self.speaker = speaker
        
        return old_attributes
    
    
    def _update_channel(self, data):
        """
        Updates the voice state's channel.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Voice state data.
        
        Returns
        -------
        old_channel_id : `int`
            The voice state's old channel's identifier.
        new_channel_id : `int`
            The voice state's new channel's identifier.
        """
        new_channel_id = parse_channel_id(data)
        if not new_channel_id:
            try:
                guild = GUILDS[self.guild_id]
            except KeyError:
                pass
            else:
                voice_states = guild.voice_states
                if (voice_states is not None):
                    try:
                        del voice_states[self.user_id]
                    except KeyError:
                        pass
                    else:
                        if not voice_states:
                            guild.voice_states = None
        
        old_channel_id = self.channel_id
        self.channel_id = new_channel_id
        
        return old_channel_id, new_channel_id
    
    
    def __repr__(self):
        """Returns the voice state's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' user_id = ')
        repr_parts.append(repr(self.user_id))
        
        repr_parts.append(', guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        repr_parts.append(', channel_id = ')
        repr_parts.append(repr(self.channel_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two voice states are equal"""
        # Shortcut
        if self is other:
            return True
        
        if type(self) is not type(other):
            return NotImplemented
        
        # channel_id
        if self.channel_id != other.channel_id:
            return False
        
        # deaf
        if self.deaf != other.deaf:
            return False
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        # mute
        if self.mute != other.mute:
            return False
        
        # requested_to_speak_at
        if self.requested_to_speak_at != other.requested_to_speak_at:
            return False
        
        # self_deaf
        if self.self_deaf != other.self_deaf:
            return False
        
        # self_mute
        if self.self_mute != other.self_mute:
            return False
        
        # self_stream
        if self.self_stream != other.self_stream:
            return False
        
        # self_video
        if self.self_video != other.self_video:
            return False
        
        # session_id
        if self.session_id != other.session_id:
            return False
        
        # speaker
        if self.speaker != other.speaker:
            return False
        
        # user_id
        if self.user_id != other.user_id:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the voice state's hash value."""
        hash_value = 0
        
        # channel_id
        hash_value ^= self.channel_id
        
        # deaf
        hash_value ^= self.deaf << 0
        
        # guild_id
        hash_value ^= self.guild_id
        # mute
        hash_value ^= self.mute << 2
        
        # requested_to_speak_at
        requested_to_speak_at = self.requested_to_speak_at
        if (requested_to_speak_at is not None):
            hash_value ^= hash(requested_to_speak_at)
        
        # self_deaf
        hash_value ^= self.self_deaf << 3
        
        # self_mute
        hash_value ^= self.self_mute << 4
        
        # self_stream
        hash_value ^= self.self_stream << 5
        
        # self_video
        hash_value ^= self.self_video << 6
        
        # session_id
        hash_value ^= hash(self.session_id)
        
        # speaker
        hash_value ^= self.speaker << 1
        
        
        # user_id
        hash_value ^= self.user_id
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the voice state.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        # Construct
        new = object.__new__(type(self))
        new._cache_user = None
        new.channel_id = self.channel_id
        new.deaf = self.deaf
        new.guild_id = self.guild_id
        new.mute = self.mute
        new.requested_to_speak_at = self.requested_to_speak_at
        new.self_deaf = self.self_deaf
        new.self_mute = self.self_mute
        new.self_stream = self.self_stream
        new.self_video = self.self_video
        new.session_id = self.session_id
        new.speaker = self.speaker
        new.user_id = self.user_id
        return new
    
    
    def copy_with(
        self,
        *,
        channel_id = ...,
        deaf = ...,
        guild_id = ...,
        speaker = ...,
        mute = ...,
        requested_to_speak_at = ...,
        self_deaf = ...,
        self_mute = ...,
        self_stream = ...,
        self_video = ...,
        session_id = ...,
        user_id = ...,
    ):
        """
        Copies the voice state with the given fields.
        
        Parameters
        ----------
        channel_id : `int`, Optional (Keyword only)
            The channel's identifier to where the user is connected to.
        deaf : `bool`, Optional (Keyword only)
            Whether the user is deafen.
        guild_id : `int`, Optional (Keyword only)
            The guild's identifier where the user is connected to.
        mute : `bool`, Optional (Keyword only)
            Whether the user is muted.
        requested_to_speak_at : `None | DateTime`, Optional (Keyword only)
            When the user requested to speak.
        self_deaf : `bool`, Optional (Keyword only)
            Whether the user muted everyone else.
        self_mute : `bool`, Optional (Keyword only)
            Whether the user muted itself.
        self_stream : `bool`, Optional (Keyword only)
            Whether the user screen shares with the go live option.
        self_video : `bool`, Optional (Keyword only)
            Whether the user sends video from a camera source.
        session_id : `str``, Optional (Keyword only)
            The user's voice session id.
        speaker : `bool`, Optional (Keyword only)
            Whether the user is suppressed inside of the voice channel.
        user_id : `int`, Optional (Keyword only)
            The voice state's respective user's identifier.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # channel_id
        if channel_id is ...:
            channel_id = self.channel_id
        else:
            channel_id = validate_channel_id(channel_id)
        
        # deaf
        if deaf is ...:
            deaf = self.deaf
        else:
            deaf = validate_deaf(deaf)
        
        # guild_id
        if guild_id is ...:
            guild_id = self.guild_id
        else:
            guild_id = validate_guild_id(guild_id)
        
        # mute
        if mute is ...:
            mute = self.mute
        else:
            mute = validate_mute(mute)
        
        # requested_to_speak_at
        if requested_to_speak_at is ...:
            requested_to_speak_at = self.requested_to_speak_at
        else:
            requested_to_speak_at = validate_requested_to_speak_at(requested_to_speak_at)
        
        # self_deaf
        if self_deaf is ...:
            self_deaf = self.self_deaf
        else:
            self_deaf = validate_self_deaf(self_deaf)
        
        # self_mute
        if self_mute is ...:
            self_mute = self.self_mute
        else:
            self_mute = validate_self_mute(self_mute)
        
        # self_stream
        if self_stream is ...:
            self_stream = self.self_stream
        else:
            self_stream = validate_self_stream(self_stream)
        
        # self_video
        if self_video is ...:
            self_video = self.self_video
        else:
            self_video = validate_self_video(self_video)
        
        # session_id
        if session_id is ...:
            session_id = self.session_id
        else:
            session_id = validate_session_id(session_id)
        
        # speaker
        if speaker is ...:
            speaker = self.speaker
        else:
            speaker = validate_speaker(speaker)
        
        # user_id
        if user_id is ...:
            user_id = self.user_id
        else:
            user_id = validate_user_id(user_id)
        
        # Construct
        new = object.__new__(type(self))
        new._cache_user = None
        new.channel_id = channel_id
        new.deaf = deaf
        new.guild_id = guild_id
        new.mute = mute
        new.requested_to_speak_at = requested_to_speak_at
        new.self_deaf = self_deaf
        new.self_mute = self_mute
        new.self_stream = self_stream
        new.self_video = self_video
        new.session_id = session_id
        new.speaker = speaker
        new.user_id = user_id
        return new
    
    
    @property
    def user(self):
        """
        Returns the voice state's user.
        
        Returns
        -------
        user : ``ClientUserBase``
        """
        user = self._cache_user
        if (user is None):
            user = create_partial_user_from_id(self.user_id)
            self._cache_user = user
        
        return user
    
    
    def _set_cache_user(self, user):
        """
        Sets the cached user of the voice state.
        
        Parameters
        ----------
        user : ``ClientUserBase``
            The respective user of the voice state.
        """
        self._cache_user = user
    
    
    @property
    def channel(self):
        """
        Returns the voice state's channel, where the user is connected to.
        
        Returns
        -------
        channel : ``Channel``
        """
        return create_partial_channel_from_id(self.channel_id, ChannelType.unknown, self.guild_id)
    
    
    @property
    def guild(self):
        """
        Returns the voice state's respective guild.
        
        Returns
        -------
        guild : ``None | Guild``
        """
        return GUILDS.get(self.guild_id, None)



__all__ = ('PermissionMismatchWarning',)

from warnings import warn

from ...discord.application import Team
from ...discord.application_command import ApplicationCommandPermissionOverwriteTargetType


class PermissionMismatchWarning(RuntimeWarning):
    """
    Warnings used when an application has different permissions than expected.
    """


def _maybe_render_application_command_permission_overwrites_to(
    title_line, application_command_permission_overwrites, into
):
    """
    Renders the given permission overwrites to the given list of strings.
    
    Parameters
    ----------
    application_command_permission_overwrites : `set` of ``ApplicationCommandPermissionOverwrite``
        The permission overwrites to show.
    into : `list` of `str`
        The list to render the representation of the application command permission overwrites to.
    
    Returns
    -------
    into : `list` of `str`
    """
    if application_command_permission_overwrites:
        into.append('\n')
        into.append(title_line)
        
        into = _render_application_command_permission_overwrites_to(application_command_permission_overwrites, into)
        
    return into


def _render_application_command_permission_overwrites_to(application_command_permission_overwrites, into):
    """
    Renders the given permission overwrites to the given list of strings.
    
    Parameters
    ----------
    application_command_permission_overwrites : `set` of ``ApplicationCommandPermissionOverwrite``
        The permission overwrites to show.
    into : `list` of `str`
        The list to render the representation of the application command permission overwrites to.
    
    Returns
    -------
    into : `list` of `str`
    """
    for index, application_command_permission_overwrite in (
        enumerate(sorted(application_command_permission_overwrites), 1)
    ):
        index_name = str(index)
        into.append('\n')
        into.append(index_name)
        into.append(':')
        into.append(' ' * max(0, 3 - len(index_name)))
        into.append('target type = ')
        into.append(application_command_permission_overwrite.target_type.name)
        into.append('\n    target id = ')
        into.append(str(application_command_permission_overwrite.target_id))
        into.append('\n    allow = ')
        into.append('true' if application_command_permission_overwrite.allow else 'false')
    
    return into


def create_permission_mismatch_message(
    application_command,
    guild_id,
    current_application_command_permission_overwrites,
    expected_application_command_permission_overwrites
):
    """
    Creates permission mismatch message.
    
    Parameters
    ----------
    application_command : ``ApplicationCommand``
        The respective application command-
    guild_id : `int`
        The respective guild's identifier where the command is.
    current_application_command_permission_overwrites : `None`, `list` of ``ApplicationCommandPermissionOverwrite``
        The actual permission overwrites of the command.
    expected_application_command_permission_overwrites : `None`, `list` of ``ApplicationCommandPermissionOverwrite``
        The expected permission overwrites of the command.
    
    Returns
    -------
    message : `str`
    """
    message_parts = []
    
    message_parts.append('\nApplication command permission mismatch at:')
    
    message_parts.append('\n    guild id = ')
    message_parts.append(str(guild_id))
    
    if (application_command is not None):
        message_parts.append('\n    command id = ')
        message_parts.append(str(application_command.id))
        
        message_parts.append('\n    command name = ')
        message_parts.append(str(application_command.name))
    
    
    if current_application_command_permission_overwrites is None:
        current_application_command_permission_overwrites = set()
    else:
        current_application_command_permission_overwrites = {*current_application_command_permission_overwrites}
    
    if expected_application_command_permission_overwrites is None:
        expected_application_command_permission_overwrites = set()
    else:
        expected_application_command_permission_overwrites = {*expected_application_command_permission_overwrites}
    
    
    message_parts = _maybe_render_application_command_permission_overwrites_to(
        'Shared application command permission overwrites:',
        current_application_command_permission_overwrites & expected_application_command_permission_overwrites,
        message_parts,
    )
    
    message_parts = _maybe_render_application_command_permission_overwrites_to(
        'Additional application command permission overwrites:',
        current_application_command_permission_overwrites - expected_application_command_permission_overwrites,
        message_parts,
    )
    
    message_parts = _maybe_render_application_command_permission_overwrites_to(
        'Missing application command permission overwrites:',
        expected_application_command_permission_overwrites - current_application_command_permission_overwrites,
        message_parts,
    )
    
    
    return ''.join(message_parts)


def check_and_warn_can_request_owners_access_of(client):
    """
    Checks whether the client's access can be requested. If it cannot drops a warning.
    
    Parameters
    ----------
    client : ``Client``
        The respective client.
    
    Returns
    -------
    can_request_owners_access : `bool`
    """
    while True:
        secret = client.secret
        if secret is None:
            failure_reason = '`client.secret` is None`.'
            break
        
        if isinstance(client.application.owner, Team):
            failure_reason = 'Client application owned by team.'
            break
        
        failure_reason = True
        break
    
    if failure_reason is None:
        return True
    
    
    warn(
        (
            f'\n'
            f'Requesting owner\'s access impossible:\n'
            f'reason: {failure_reason}\n'
            f'client: {client!r}'
        ),
        PermissionMismatchWarning,
    )
    return False


def _reduce_application_command_permission_overwrites(guild_id, application_command_permission_overwrites):
    """
    Removes the application command permission overwrites which do nothing at all.
    
    Parameters
    ----------
    guild_id : `int`
        The respective guild's identifier where the command is.
    application_command_permission_overwrites : `None`, `list` of ``ApplicationCommandPermissionOverwrite``
        The application command permission overwrites to reduce.
    
    Returns
    -------
    reduced : `None`, `list` of ``ApplicationCommandPermissionOverwrite``
    """
    # Optimal case
    if application_command_permission_overwrites is None:
        return None
    
    # Group up
    default_role_permission_overwrite = None
    mentionable_permission_overwrites = None
    default_channel_permission_overwrite = None
    channel_permission_overwrites = None
    unknown_permission_overwrites = None
    
    for application_command_permission_overwrite in application_command_permission_overwrites:
        target_type = application_command_permission_overwrite.target_type
        if target_type is ApplicationCommandPermissionOverwriteTargetType.user:
            if mentionable_permission_overwrites is None:
                mentionable_permission_overwrites = []
            
            mentionable_permission_overwrites.append(application_command_permission_overwrite)
        
        elif target_type is ApplicationCommandPermissionOverwriteTargetType.role:
            if application_command_permission_overwrite.target_id == 0:
                default_role_permission_overwrite = application_command_permission_overwrite
            
            elif application_command_permission_overwrite.target_id == guild_id:
                application_command_permission_overwrite = application_command_permission_overwrite.copy_with(
                    target = ('role', 0)
                )
                
                default_role_permission_overwrite = application_command_permission_overwrite
            
            else:
                if mentionable_permission_overwrites is None:
                    mentionable_permission_overwrites = []
                
                mentionable_permission_overwrites.append(application_command_permission_overwrite)
        
        elif target_type is ApplicationCommandPermissionOverwriteTargetType.channel:
            if application_command_permission_overwrite.target_id == 0:
                default_channel_permission_overwrite = application_command_permission_overwrite
            
            elif application_command_permission_overwrite.target_id == guild_id - 1:
                application_command_permission_overwrite = application_command_permission_overwrite.copy_with(
                    target = ('channel', 0)
                )
                default_channel_permission_overwrite = application_command_permission_overwrite
            
            else:
                if channel_permission_overwrites is None:
                    channel_permission_overwrites = []
                
                channel_permission_overwrites.append(application_command_permission_overwrite)
        
        else:
            if unknown_permission_overwrites is None:
                unknown_permission_overwrites = []
            
            unknown_permission_overwrites.append(application_command_permission_overwrite)
    
    # process
    
    reduced = []
    
    if default_role_permission_overwrite is None:
        default_role_allow = True
    elif default_role_permission_overwrite.allow:
        default_role_allow = True
    else:
        default_role_allow = False
        
        reduced.append(default_role_permission_overwrite)
    
    
    if default_channel_permission_overwrite is None:
        default_channel_allow = True
    elif default_channel_permission_overwrite.allow:
        default_channel_allow = True
    else:
        default_channel_allow = False
        
        reduced.append(default_channel_permission_overwrite)
    
    
    if (mentionable_permission_overwrites is not None):
        for mentionable_permission_overwrite in mentionable_permission_overwrites:
            if mentionable_permission_overwrite.allow != default_role_allow:
                reduced.append(mentionable_permission_overwrite)
    
    if (channel_permission_overwrites is not None):
        for channel_permission_overwrite in channel_permission_overwrites:
            if channel_permission_overwrite.allow != default_channel_allow:
                reduced.append(channel_permission_overwrite)
    
    
    if (unknown_permission_overwrites is not None):
        reduced.extend(unknown_permission_overwrites)
    
    if not reduced:
        return None
        
    reduced.sort()
    return reduced


def are_application_command_permission_overwrites_equal(
    guild_id,
    current_application_command_permission_overwrites,
    expected_application_command_permission_overwrites,
):
    """
    Returns whether the two application command permission overwrites are equal.
    
    Parameters
    ----------
    guild_id : `int`
        The respective guild's identifier where the command is.
    current_application_command_permission_overwrites : `None`, `list` of ``ApplicationCommandPermissionOverwrite``
        The actual permission overwrites of the command.
    expected_application_command_permission_overwrites : `None`, `list` of ``ApplicationCommandPermissionOverwrite``
        The expected permission overwrites of the command.
    
    Returns
    -------
    are_equal : `bool`
    """
    if (
        (current_application_command_permission_overwrites is None) and
        (expected_application_command_permission_overwrites is None)
    ):
        return True
    
    current_application_command_permission_overwrites = _reduce_application_command_permission_overwrites(
        guild_id, current_application_command_permission_overwrites 
    )
    
    expected_application_command_permission_overwrites = _reduce_application_command_permission_overwrites(
        guild_id, expected_application_command_permission_overwrites
    )
    
    if current_application_command_permission_overwrites == expected_application_command_permission_overwrites:
        return True
    
    return False



__all__ = ('Slotted', )

from itertools import islice
from types import MemberDescriptorType


def _get_direct_parent(meta_type, type_name, type_parents):
    """
    Gets the direct parent of the specified type definition and it's merged slots.
    
    Parameters
    ----------
    type_name : `str`
        The created type's name.
    
    type_parents : `tuple<type>`
        The parent types of the creates type.
    
    type_attributes : `dict<str, object>`
        The type attributes of the created type.
    
    Returns
    -------
    direct_parent : `None | type`
        The direct parent of the respective type.
    
    Raises
    ------
    RuntimeError
    """
    parent_count = len(type_parents)
    if parent_count > 0:
        direct_parent = type_parents[0]
    else:
        direct_parent = None
    
    for type_parent in islice(type_parents, 1, None):
        if isinstance(type_parent, meta_type):
            raise RuntimeError(
                f'`{type_name}` wanted to inherit a `{meta_type.__name__}` type not as it\'s direct '
                f'(0th) parent type.'
            )
    
    return direct_parent


def _inherit_hash(direct_parent, type_attributes):
    """
    Inherits the hasher into the type attributes because someone decided to pop it.
    
    Parameters
    ----------
    direct_parent : `None | type`
        The direct parent of the respective type.
    
    type_attributes : `dict<str, object>`
        The type attributes of the created type.
    """
    # Sub-types might miss hash, lmeow
    if (direct_parent is not None) and type_attributes.get('__hash__', None) is None:
        type_attributes['__hash__'] = direct_parent.__hash__


def _merge_type_slots(direct_parent, type_parents, type_attributes):
    """
    Merges the new values into the slots of the type.
    
    Parameters
    ----------
    direct_parent : `None | type`
        The direct parent of the respective type.
    
    type_parents : `tuple<type>`
        The parent types of the creates type.
    
    type_attributes : `dict<str, object>`
        The type attributes of the created type.
    
    Returns
    -------
    final_slots : `set<str>`
        Slots for the created type.
    """
    final_slots = set()
    
    if (direct_parent is not None):
        for type_parent in type_parents:
            for slot_name in getattr(type_parent, f'_{type_parent.__name__}__slots', ()):
                existing_attribute = getattr(direct_parent, slot_name, None)
                if (existing_attribute is not None) and isinstance(existing_attribute, MemberDescriptorType):
                    continue
                
                final_slots.add(slot_name)
    
    slots = type_attributes.get('__slots__', None)
    if (slots is not None) and slots:
        final_slots.update(slots)
    
    # Remove weakref if we already have it
    if (direct_parent is not None) and hasattr(direct_parent, '__weakref__'):
        try:
            final_slots.remove('__weakref__')
        except KeyError:
            pass
    
    return final_slots


def _process_set_slot(type_attributes, final_slots):
    """
    Processes type attributes, which implement  `__set_slot__`.
    
    Parameters
    ----------
    type_attributes : `dict<str, object>`
        The type attributes of the created type.
    
    final_slots : `set<str>`
        Slots for the created type.
    """
    slotters = []
    for attribute_item in type_attributes.items():
        attribute_value = attribute_item[1]
        
        # Ignore types.
        if isinstance(attribute_value, type):
            continue
        
        # Check the type, whether it has the correct method.
        set_slot = getattr(type(attribute_value), '__set_slot__', None)
        if set_slot is None:
            continue
        
        # Queue up for future applications.
        slotters.append(attribute_item)
        continue
    
    # Apply slotters
    while slotters:
        attribute_name, slotter = slotters.pop()
        type(slotter).__set_slot__(slotter, attribute_name, type_attributes, final_slots)


class Slotted(type):
    """
    Meta type for special slotted objects, which require `__set_slot__`.
    """
    def __new__(cls, type_name, type_parents, type_attributes):
        """
        Creates a new ``Slotted``.
        
        Parameters
        ----------
        type_name : `str`
            The created type's name.
        
        type_parents : `tuple<type>`
            The sub-types of the creates type.
        
        type_attributes : `dict<str, object>`
            The type attributes of the created type.
        """
        direct_parent = _get_direct_parent(cls, type_name, type_parents)
        final_slots = _merge_type_slots(direct_parent, type_parents, type_attributes)
        _process_set_slot(type_attributes, final_slots)
        _inherit_hash(direct_parent, type_attributes)
        
        type_attributes['__slots__'] = tuple(sorted(final_slots))
        
        return type.__new__(cls, type_name, type_parents, type_attributes)



__all__ = ()

from ..reaction_mapping_line import ReactionMappingLine

from ..emoji import Emoji
from ..parsing import parse_reaction
from ..reaction import Reaction, ReactionType


def validate_reaction(reaction):
    """
    Validates whether the given reaction value.
    
    Parameters
    ----------
    reaction : ``Emoji``, ``Reaction``, `str`
        The reaction to validate.
    
    Returns
    -------
    reaction : ``Reaction``
    
    Raises
    ------
    TypeError
        - If `reaction`'s type is incorrect.
    """
    if isinstance(reaction, Reaction):
        return reaction
    
    if isinstance(reaction, Emoji):
        return Reaction.from_fields(reaction, ReactionType.standard)
    
    if isinstance(reaction, str):
        emoji = parse_reaction(reaction)
        if emoji is not None:
            return Reaction.from_fields(emoji, ReactionType.standard)
        
        raise ValueError(
            f'`reaction` is given as `str`, but could not parse it. Got {reaction!r}.'
        )
    
    raise TypeError(
        f'Reaction can be `{Reaction.__name__}`, `{Emoji.__name__}`, `str`, '
        f'got {reaction.__class__.__name__}; {reaction!r}.'
    )


def validate_reaction_mapping_line(reaction_mapping_line):
    """
    Validates the given `reaction_mapping_line` value.
    
    Parameters
    ----------
    reaction_mapping_line : ``ReactionMappingLine``
        The reaction mapping line to validate.
    
    Returns
    -------
    reaction_mapping_line : ``ReactionMappingLine``
    
    Raises
    ------
    TypeError
        - If `reaction_mapping_line`'s type is incorrect.
    """
    if not isinstance(reaction_mapping_line, ReactionMappingLine):
        raise TypeError(
            f'Reaction mapping line can be can be `{ReactionMappingLine.__name__}`, '
            f'got {type(reaction_mapping_line).__name__}; {reaction_mapping_line!r}.'
        )
    
    return reaction_mapping_line


def validate_lines(lines):
    """
    Validates a reaction mapping's lines.
    
    Parameters
    ----------
    lines : `None | dict<str | Emoji | Reaction, ReactionMappingLine> \
            | list<(str | Emoji | Reaction, ReactionMappingLine)>`
        The value to initialise the reaction mapping.
    
    Returns
    -------
    built_lines : `None`, `list` of `tuple` ((``Emoji``, ``Reaction``), ``ReactionMappingLine``) items
        The validated extend with value.
    
    Raises
    ------
    TypeError
        - If `lines`'s type is unacceptable.
        - If an item (or element) of `lines` has incorrect type, length or structure.
    """
    if lines is None:
        built_lines = None
    
    elif isinstance(lines, dict):
        built_lines = None
        for item in lines.items():
            built_lines = _validate_reaction_mapping_lines_item(built_lines, item)
    
    elif (getattr(lines, '__iter__', None) is not None):
        built_lines = None
        
        for item in lines:
            if not isinstance(item, tuple):
                raise TypeError(
                    f'`lines` items can be `tuple` instances, got '
                    f'{item.__class__.__name__}; {item!r}; lines = {lines!r}.'
                )
            
            item_length = len(item)
            if len(item) != 2:
                raise TypeError(
                    f'`lines` items can be `tuple` with length of `2`, got '
                    f'item_length = {item_length!r}; item = {item!r}; lines = {lines!r}.'
                )
            
            built_lines = _validate_reaction_mapping_lines_item(built_lines, item)
        
    else:
        raise TypeError(
            f'`lines` can be `None`, `iterable` or `dict`, got '
            f'{type(lines).__name__}; {lines!r}'
        )
    
    return built_lines


def _validate_reaction_mapping_lines_item(built_lines, item):
    """
    Validates an item of the `lines` parameter of ``ReactionMapping``.
    
    Parameters
    ----------
    built_lines : `None | dict<Reaction, ReactionMappingLine>`
        The validated extend with value.
    item : `(str | Emoji | Reaction, ReactionMappingLine)`
        Reaction mapping item to validate.
    
    Returns
    -------
    built_lines : `None | dict<Reaction, ReactionMappingLine>`
        The validated extend with value.
    
    Raises
    ------
    TypeError
        - If `item[0]` is not `str` / ``Emoji`` / ``Reaction``.
        - If `item[1]` is not ``ReactionMappingLine``.
    """
    reaction, line = item
    
    reaction = validate_reaction(reaction)
    line = validate_reaction_mapping_line(line)
    
    if line.count:
        if built_lines is None:
            built_lines = {}
        
        built_lines[reaction] = line
    
    return built_lines



from .audit_log_entry import *
from .constants import *
from .fields import *
from .preinstanced import *
from .target_converters import *


__all__ = (
    *audit_log_entry.__all__,
    *constants.__all__,
    *fields.__all__,
    *preinstanced.__all__,
    *target_converters.__all__,
)


from ....utils.module_deprecation import deprecated_import
deprecated_import(AuditLogEntryType, 'AuditLogEvent')
deprecated_import(AuditLogEntryTargetType, 'AuditLogTargetType')



__all__ = ()

from ...field_parsers import nullable_string_parser_factory
from ...field_putters import nullable_string_putter_factory
from ...field_validators import nullable_string_validator_factory

from .constants import (
    PLATFORM_NAME_LENGTH_MAX, PLATFORM_NAME_LENGTH_MIN, PLATFORM_USER_NAME_LENGTH_MAX, PLATFORM_USER_NAME_LENGTH_MIN
)

# platform_name

parse_platform_name = nullable_string_parser_factory('platform_name')
put_platform_name = nullable_string_putter_factory('platform_name')
validate_platform_name = nullable_string_validator_factory(
    'platform_name', PLATFORM_NAME_LENGTH_MIN, PLATFORM_NAME_LENGTH_MAX
)


# platform_user_name

parse_platform_user_name = nullable_string_parser_factory('platform_username')
put_platform_user_name = nullable_string_putter_factory('platform_username')
validate_platform_user_name = nullable_string_validator_factory(
    'platform_username', PLATFORM_USER_NAME_LENGTH_MIN, PLATFORM_USER_NAME_LENGTH_MAX
)

# metadata_values

def parse_metadata_values(data):
    """
    Parsers out metadata values from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Application role connection data.
    
    Returns
    -------
    metadata_values : `None | dict<str, str>`
    """
    metadata_values = data.get('metadata', None)
    if (metadata_values is not None) and metadata_values:
        return metadata_values


def put_metadata_values(metadata_values, data, defaults):
    """
    Puts the application role connection metadata values into the given data.
    
    Parameters
    ----------
    metadata_values : `None | dict<str, str>`
        Metadata values.
    data : `dict<str, object>`
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if defaults or (metadata_values is not None):
        if metadata_values is None:
            metadata_values = {}
        
        data['metadata'] = metadata_values
    
    return data


def validate_metadata_values(metadata_values):
    """
    Validates the given application role connection metadata values.
    
    Parameters
    ----------
    metadata_values : `None | dict<str, str>`
        Metadata values.
    
    Returns
    -------
    metadata_values : `None | dict<str, str>`
        Metadata values.
    
    TypeError
        - If `metadata_values`'s type or structure is incorrect.
    """
    if metadata_values is None:
        return None
    
    if not isinstance(metadata_values, dict):
        raise TypeError(
            f'`metadata_values` can be `dict`, got {metadata_values.__class__.__name__}; {metadata_values!r}.'
        )
    
    if not metadata_values:
        return None
    
    for key, value in metadata_values.items():
        if not isinstance(key, str):
            raise TypeError(
                f'`metadata_values` keys can be `str`, got {key.__class__.__name__}; {key!r}, '
                f'metadata_values = {metadata_values!r}.'
            )
        
        if not isinstance(value, str):
            raise TypeError(
                f'`metadata_values` values can be `str`, got {value.__class__.__name__}; {value!r}, '
                f'metadata_values = {metadata_values!r}.'
            )
    
    return metadata_values



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, force_string_parser_factory, int_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_optional_putter_factory, force_string_putter_factory, int_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_validator_factory, force_string_validator_factory,
    int_conditional_validator_factory
)


# renewal

parse_renewal = bool_parser_factory('is_renewal', False)
put_renewal = bool_optional_putter_factory('is_renewal', False)
validate_renewal = bool_validator_factory('renewal', False)

# subscription_listing_id

parse_subscription_listing_id = entity_id_parser_factory('role_subscription_listing_id')
put_subscription_listing_id = entity_id_optional_putter_factory('role_subscription_listing_id')
validate_subscription_listing_id = entity_id_validator_factory('subscription_listing_id')

# tier_name

parse_tier_name = force_string_parser_factory('tier_name')
put_tier_name = force_string_putter_factory('tier_name')
validate_tier_name = force_string_validator_factory('tier_name', 0, 1024)

# total_months

parse_total_months = int_parser_factory('total_months_subscribed', 1)
put_total_months = int_putter_factory('total_months_subscribed')
validate_total_months = int_conditional_validator_factory(
    'total_months',
    1,
    lambda total_months : total_months >= 1,
    '>= 1',
)
